from relationalai.early_access.dsl.core.types.constrained.subtype import ValueSubtype
from relationalai.early_access.dsl.core.types.unconstrained import UnconstrainedValueType, UnconstrainedNumericType

standard_value_types = {}

# Hash types
Hash = UnconstrainedValueType('Hash')
standard_value_types[Hash.display()] = Hash

# Boolean types
Boolean = UnconstrainedNumericType('Boolean')
standard_value_types[Boolean.display()] = Boolean

# Decimal types
#
Decimal = UnconstrainedNumericType('Decimal')
standard_value_types[Decimal.display()] = Decimal

UnsignedDecimal = ValueSubtype('UnsignedDecimal')
standard_value_types[UnsignedDecimal.display()] = UnsignedDecimal
with UnsignedDecimal:
    x = Decimal()
    x >= 0

PositiveDecimal = ValueSubtype('PositiveDecimal')
standard_value_types[PositiveDecimal.display()] = PositiveDecimal
with PositiveDecimal:
    x = UnsignedDecimal()
    x != 0

# Floating-point types
#
Float = UnconstrainedNumericType('Float')
standard_value_types[Float.display()] = Float
Double = UnconstrainedNumericType('Double')  # A float64 in Rel
standard_value_types[Double.display()] = Double

# Integer types
#
BigInteger = UnconstrainedNumericType('BigInteger')  # 128-bit integer
standard_value_types[BigInteger.display()] = BigInteger

BigUnsignedInteger = ValueSubtype('BigUnsignedInteger')
standard_value_types[BigUnsignedInteger.display()] = BigUnsignedInteger
with BigUnsignedInteger:
    x = BigInteger()
    x >= 0

BigPositiveInteger = ValueSubtype('BigPositiveInteger')
standard_value_types[BigPositiveInteger.display()] = BigPositiveInteger
with BigPositiveInteger:
    x = BigUnsignedInteger()
    x != 0

Integer = UnconstrainedNumericType('Integer')
standard_value_types[Integer.display()] = Integer

UnsignedInteger = ValueSubtype('UnsignedInteger')
standard_value_types[UnsignedInteger.display()] = UnsignedInteger
with UnsignedInteger:
    x = Integer()
    x >= 0

PositiveInteger = ValueSubtype('PositiveInteger')
standard_value_types[PositiveInteger.display()] = PositiveInteger
with PositiveInteger:
    x = UnsignedInteger()
    x != 0

# Date types
#
Date = UnconstrainedValueType('Date')
standard_value_types[Date.display()] = Date

# DateTime types
#
DateTime = UnconstrainedValueType('DateTime')
standard_value_types[DateTime.display()] = DateTime

RowId = UnconstrainedNumericType('RowId')
standard_value_types[RowId.display()] = RowId

# String types
#
String = UnconstrainedValueType('String')
standard_value_types[String.display()] = String

Any = UnconstrainedValueType('Any')
standard_value_types[Any.display()] = Any

Symbol = UnconstrainedValueType('Symbol')
standard_value_types[Symbol.display()] = Symbol



from relationalai.semantics.sql.rewrite.denormalize import Denormalize
from relationalai.semantics.sql.rewrite.recursive_union import RecursiveUnion
from relationalai.semantics.sql.rewrite.double_negation import DoubleNegation
from relationalai.semantics.sql.rewrite.sort_output_query import SortOutputQuery

__all__ = ["Denormalize", "RecursiveUnion", "DoubleNegation", "SortOutputQuery"]



import dataclasses
from typing import Optional

from relationalai.early_access.dsl.core.namespaces import Namespace
from relationalai.early_access.dsl.core.relations import Relation, RelationSignature
from relationalai.early_access.dsl.core.types import Type
from relationalai.early_access.dsl.core.utils import generate_stable_uuid
from relationalai.early_access.dsl.ontologies.relationships import Relationship
from relationalai.early_access.dsl.ontologies.roles import Role
from relationalai.early_access.dsl.physical_metadata.tables import Column, Table, TemporalTable
from relationalai.early_access.dsl.types.entities import EntityType
from relationalai.early_access.dsl.types.values import ValueType

from relationalai.early_access.dsl import ExternalRelation


@dataclasses.dataclass
class EntityMapRef:
    name: str
    source: Table
    entity_type: EntityType
    relation: Relation


@dataclasses.dataclass
class ReificationPattern:
    concept: type
    reifies_as: type


class Binding:

    def __init__(self, role: Role, column: Optional[Column] = None, entity_map: Optional[EntityMapRef] = None):
        self._role = role
        self._column = column
        self._entity_map = entity_map
        self._has_value: Optional[str] = None
        self._applied_transformations = []
        self._check_either_column_or_entity_map_provided()

    def guid(self):
        salt = self._column.guid() if self._column is not None else self._entity_map.name if self._entity_map is not None else None
        return generate_stable_uuid(f"{salt}_{self._role.guid()}")

    def transform(self, transformer: Relation):
        self._applied_transformations.append(transformer)
        return self

    @property
    def role(self):
        return self._role

    @property
    def column(self):
        return self._column

    @property
    def entity_map(self):
        return self._entity_map

    @property
    def has_value(self) -> Optional[str]:
        return self._has_value

    @has_value.setter
    def has_value(self, value: str):
        if value is not None:
            self._has_value = value
            Binding._check_has_value_applicable(self._role.part_of)

    def __str__(self):
        binds_to = self._column.pprint() if self._column is not None else self._entity_map.name if self._entity_map is not None else None
        return f"{self._role.verbalize()}: {binds_to}"

    def _check_either_column_or_entity_map_provided(self):
        if self._column is None and self._entity_map is None:
            raise Exception("Either a column or an entity map must be provided")

    @staticmethod
    def _check_has_value_applicable(relationship):
        if not (relationship.arity() == 1 or relationship.is_subtype()):
            raise Exception(
                f"The 'has_value' setting requires a unary or a subtype relationship, "
                f"but non-subtype relationship '{relationship.name()}' has arity {relationship.arity()}")


class BindingModel:
    def __init__(self):
        self._bindings = []
        self._patterns = {}
        self._transformers = {}
        self._entity_maps = {}

    def bind_role(self, role: Role, column: Optional[Column] = None, entity_map: Optional[EntityMapRef] = None,
                  data_transformers: Optional[list[Relation]] = None, has_value: Optional[str] = None):
        if column is not None and isinstance(column.part_of, TemporalTable):
            if column.part_of.temporal_col.name not in column.part_of.columns:
                raise Exception(f"Cannot bind attribute {column.name} of TemporalTable {column.part_of.name} as it uses"
                                f" missing temporal column: column {column.part_of.temporal_col} is not in the table")
        b = Binding(role, column, entity_map)
        self._register_transformations(b, data_transformers)
        b._has_value = has_value
        self._bindings.append(b)
        return b

    def bind_attribute(self, relationship: Relationship, column: Column,
                       data_transformers: Optional[list[Relation]] = None, has_value: Optional[str] = None):
        b = Binding(relationship.attr(), column)
        self._register_transformations(b, data_transformers)
        b._has_value = has_value
        self._bindings.append(b)
        return b

    def transformer(self, name: str, sig: tuple[Type, Type]):
        self._transformers[name] = sig
        signature = RelationSignature(*sig)
        t = ExternalRelation(Namespace.top, name, signature)
        if len(sig) < 2:
            raise Exception(f"Signature \"{sig}\" provided for the relation {name} used as transformer must be exactly"
                            f" a tuple of two types (source type [0] and target type [1])")
        return t

    def entity_map(self, name: str, source: Table, entity_type: EntityType, relation: Relation):
        emap = EntityMapRef(name, source, entity_type, relation)
        if name in self._entity_maps:
            raise Exception(f"Entity map reference with name {name} already declared")
        self._entity_maps[name] = emap
        return emap

    def pattern(self, concept: type, reifies_as: type):
        if type is not EntityType or type is not ValueType:
            raise Exception(f"Reification pattern must use EntityType or ValueType, but for {concept} got {reifies_as}")
        self._patterns[concept] = reifies_as
        return ReificationPattern(concept, reifies_as)

    def pprint(self):
        return "\n".join([str(b) for b in self._bindings])

    @staticmethod
    def _register_transformations(b: Binding, *transformations):
        if transformations is not None:
            for t in transformations:
                b.transform(t)



from __future__ import annotations

from typing import Optional, Sequence as PySequence, cast

from relationalai.early_access.metamodel import ir, compiler as c, factory as f, util
from relationalai.early_access.metamodel.helpers import collect_implicit_vars

class Splinter(c.Pass):
    """
    Splits multi-headed rules into multiple rules. Additionally, infers missing Exists tasks.
    """

    def __init__(self):
        super().__init__()
        self.name_cache = util.NameCache(start_from_one=True)

    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        if isinstance(model.root, ir.Logical):
            final = []
            new_relations:list[ir.Relation] = []
            new_relations.extend(model.relations)
            for child in model.root.body:
                new_logicals, relation = self.split(cast(ir.Logical, child))
                final.extend(new_logicals)
                if relation:
                    new_relations.append(relation)
            return ir.Model(
                    model.engines,
                    util.FrozenOrderedSet.from_iterable(new_relations),
                    model.types,
                    ir.Logical(
                        model.root.engine,
                        model.root.hoisted,
                        tuple(final)
                    )
                )
        return model

    def split(self, node: ir.Logical) -> tuple[list[ir.Logical], Optional[ir.Relation]]:
        # Split this logical, which represents a rule, into potentially many logicals, one
        # for each head (update or output)
        effects, body = self.split_items(node.body)

        if len(effects) > 1:
            effects_vars = collect_implicit_vars(*effects)
            final:list[ir.Logical] = []
            connection = None
            if body:
                # if the node has a body, create a connection, derive the body into the
                # connection, and then lookup this connection on the other logicals
                name = self.name_cache.get_name(node.id, "_intermediate")
                connection = f.relation(name, [f.field(f"f_{i+1}", v.type) for i, v in enumerate(effects_vars)])
                final.append(f.logical([*body, f.derive(connection, list(effects_vars))]))

            for effect in effects:
                if connection:
                    # there's a connection, create a logical that looks up from the connection
                    # and derives into the effect
                    lookup = f.lookup(connection, effects_vars.get_list())
                    final.append(f.logical([lookup, effect]))
                else:
                    # if there's no connection, the effect is self contained (probably literals)
                    final.append(f.logical([effect]))
            return final, connection
        return [node], None


    def split_items(self, items: PySequence[ir.Task]) -> tuple[list[ir.Task], list[ir.Task]]:
        effects = []
        body = []
        for item in items:
            if isinstance(item, (ir.Update, ir.Output)):
                effects.append(item)
            else:
                body.append(item)
        return effects, body



from .splinter import Splinter
from .list_types import RewriteListTypes
from .gc_nodes import GarbageCollectNodes
from .flatten import Flatten
from .extract_keys import ExtractKeys
from .extract_nested_logicals import ExtractNestedLogicals

__all__ = ["Splinter", "RewriteListTypes", "GarbageCollectNodes", "Flatten", "ExtractKeys", "ExtractNestedLogicals"]



"""
A simple metamodel for SQL.
"""
from __future__ import annotations

import datetime
from dataclasses import dataclass
from io import StringIO
from typing import Optional, Union, Tuple, Any
from relationalai.early_access.metamodel.util import Printer as BasePrinter
import json

@dataclass(frozen=True)
class Node:
    def __str__(self):
        return to_string(self)

@dataclass(frozen=True)
class Program(Node):
    """ The root node of the SQL program. """
    statements: list[Node]

#-------------------------------------------------
# Table
#-------------------------------------------------

@dataclass(frozen=True)
class Column(Node):
    name: str
    type: str

@dataclass(frozen=True)
class Table(Node):
    name: str
    columns: list[Column]

#-------------------------------------------------
# Statements
#-------------------------------------------------

@dataclass(frozen=True)
class CreateTable(Node):
    table: Table
    if_not_exists: bool = False

@dataclass(frozen=True)
class CreateView(Node):
    name: str
    query: Union[Select, CTE]

@dataclass(frozen=True)
class Insert(Node):
    table: str
    columns: list[str]
    values: list[Tuple[Any, ...]]
    select: Optional[Select]

@dataclass(frozen=True)
class Update(Node):
    table: str
    set: list[UpdateSet]
    where: Optional[Where]

@dataclass(frozen=True)
class UpdateSet(Node):
    name: str
    expression: str

@dataclass(frozen=True)
class Select(Node):
    distinct: bool
    vars: list[Union[VarRef, RowNumberVar, int]]
    froms: Union[list[From], Select]
    where: Optional[Where] = None
    joins: Optional[list[Join]] = None
    group_by: Optional[list[VarRef]] = None
    limit: Optional[int] = None
    is_output: bool = False

@dataclass(frozen=True)
class VarRef(Node):
    name: str
    column: Optional[str] = None
    alias: Optional[str] = None

@dataclass(frozen=True)
class From(Node):
    table: str
    alias: Optional[str]

@dataclass(frozen=True)
class Join(Node):
    table: str
    alias: Optional[str]
    on: Optional[Expr] = None

@dataclass(frozen=True)
class LeftOuterJoin(Join):
    pass

@dataclass(frozen=True)
class FullOuterJoin(Join):
    pass

@dataclass(frozen=True)
class RowNumberVar(Node):
    order_by_vars: list[OrderByVar]
    partition_by_vars: list[str]
    alias: str

@dataclass(frozen=True)
class OrderByVar(Node):
    var: str
    is_ascending: bool

# TODO: consider removing Where and make Select.where: Expr
@dataclass(frozen=True)
class Where(Node):
    expression: Expr

@dataclass(frozen=True)
class RawSource(Node):
    """ A raw source for SQL, used to represent a raw SQL query or expression. """
    src: str

#-------------------------------------------------
# Clauses
#-------------------------------------------------
# TODO: move clauses from other sections

@dataclass(frozen=True)
class CTE(Node):
    """ Common Table Expressions. """
    recursive: bool
    name: str
    columns: list[str]
    selects: list[Select]

#-------------------------------------------------
# Expressions
#-------------------------------------------------

@dataclass(frozen=True)
class Expr(Node):
    pass

@dataclass(frozen=True)
class And(Expr):
    expr: list[Expr]

@dataclass(frozen=True)
class Or(Expr):
    expr: list[Expr]

@dataclass(frozen=True)
class Terminal(Expr):
    """ Avoid going deeper in the meta-model, this is an arbitrary terminal expression."""
    expr: str

@dataclass(frozen=True)
class NotNull(Expr):
    var: str

@dataclass(frozen=True)
class NotExists(Expr):
    expr: Select

@dataclass(frozen=True)
class Like(Expr):
    left: str
    right: str

#--------------------------------------------------
# Printer
#--------------------------------------------------

def to_string(node) -> str:
    io = StringIO()
    Printer(io).print_node(node, 0)
    return io.getvalue()

@dataclass(frozen=True)
class Printer(BasePrinter):

    _RESERVED_NAMES: frozenset[str] = frozenset({"right", "left", "order"})

    def _is_reserved_name(self, name: str) -> bool:
        return isinstance(name, str) and name.lower() in self._RESERVED_NAMES

    def _join(self, args, sep=', ', indent=0, is_output=False):
        for i, s in enumerate(args):
            if i != 0:
                if indent != 0:
                    self._print(sep.rstrip())
                    self._nl()
                else:
                    self._print(sep)
            self.print_node(s, indent, is_output=is_output)

    def _print_value(self, value, quote_strings:bool = False):
        if isinstance(value, tuple):
            for i, v in enumerate(value):
                if i != 0:
                    self._print(", ")
                self._print_value(v, True)
        elif isinstance(value, str):
            if quote_strings:
                self._print(json.dumps(value))
            else:
                self._print(value)
        elif isinstance(value, bool):
            self._print(str(value).lower())
        elif isinstance(value, datetime.date):
            self._print(f"'{value}'")
        elif isinstance(value, type(None)):
            self._print("NULL")
        else:
            self._print(str(value))

    def _get_table_name(self, name: str) -> str:
        return f'"{name}"' if self._is_reserved_name(name) else name

    def print_node(self, node:Node, indent=0, inlined=False, is_output=False) -> None:
        # Table
        if isinstance(node, Program):
            for idx, d in enumerate(node.statements):
                self.print_node(d, indent)
                # Avoid an extra newline at the end of the file.
                if idx != len(node.statements) - 1:
                    self._nl()
        elif isinstance(node, Table):
            self._print(f"{self._get_table_name(node.name)} (")
            self._join(node.columns)
            self._print(")")
        elif isinstance(node, Column):
            self._print(f"{node.name} {node.type}")

        # Statements
        elif isinstance(node, CreateTable):
            clause = "IF NOT EXISTS " if node.if_not_exists else ""
            self._print(f"CREATE TABLE {clause}{node.table};")
        elif isinstance(node, CreateView):
            # TODO - crying a bit inside :(
            self._print_nl(f"DROP TABLE IF EXISTS {self._get_table_name(node.name)};")
            self._print(f"CREATE VIEW {self._get_table_name(node.name)} AS ")
            self.print_node(node.query, indent, True)
            self._print(";")
        elif isinstance(node, Insert):
            self._print(f"INSERT INTO {self._get_table_name(node.table)} ")
            if len(node.columns) > 0:
                self._print("(")
                self._join(node.columns)
                self._print(") ")
            if len(node.values) > 0:
                # We need to use `SELECT` with `UNION ALL` instead of `VALUES` because Snowflake parses and restricts
                #   certain expressions in VALUES(...). Built-in functions like HASH() or MD5() are often rejected unless used in SELECT.
                for i, value in enumerate(node.values):
                    if i != 0:
                        self._print(" UNION ALL ")
                    self._print("SELECT ")
                    self._join(value)
            if node.select is not None:
                self.print_node(node.select, indent, True)
            self._print(";")
        elif isinstance(node, Update):
            self._print(f"UPDATE {self._get_table_name(node.table)} SET ")
            self._join(node.set)
            if node.where is not None:
                self.print_node(node.where, indent)
            self._print(";")
        elif isinstance(node, UpdateSet):
            self._print(f"{node.name} = {node.expression}")
        elif isinstance(node, Select):
            self._indent_print(indent, "SELECT ")
            if node.distinct and node.froms:
                self._print("DISTINCT ")
            self._join(node.vars, is_output=node.is_output or is_output)
            if node.froms:
                self._print(" FROM ")
                if isinstance(node.froms, Select):
                    # If `froms` is a `Select`, we need to print it inline
                    self._print("( ")
                    self.print_node(node.froms, indent, True)
                    self._print(" )")
                else:
                    self._join(node.froms)
            if node.joins:
                self._print(" ")
                self._join(node.joins, sep=" ")
            if node.where:
                self.print_node(node.where, indent)
            if node.group_by:
                self._print(" GROUP BY ")
                self._join(node.group_by)
            if node.limit:
                self._print(f" LIMIT {node.limit}")
            if not inlined:
                self._print(";")
        elif isinstance(node, VarRef):
            if node.column is None:
                self._print(node.name)
            else:
                if node.column.lower() == "any":
                    self._print(f'{node.name}."{node.column}"')
                else:
                    self._print(f"{node.name}.{node.column}")
            if node.alias:
                if is_output:
                    self._print(f' as "{node.alias}"')
                elif node.alias != node.column:
                    if any(c in node.alias for c in ['-', '?']):
                        self._print(f' as "{node.alias}"')
                    else:
                        self._print(f" as {node.alias}")
        elif isinstance(node, RowNumberVar):
            self._print("ROW_NUMBER() OVER (")
            if node.partition_by_vars:
                self._print(" PARTITION BY ")
                self._join(node.partition_by_vars)
            self._print(" ORDER BY ")
            self._join(node.order_by_vars)
            if is_output:
                self._print(f' ) as "{node.alias}"')
            else:
                self._print(f' ) as {node.alias}')
        elif isinstance(node, OrderByVar):
            self._print(node.var)
            if node.is_ascending:
                self._print(" ASC")
            else:
                self._print(" DESC")
        elif isinstance(node, From):
            self._print(self._get_table_name(node.table))
            if node.alias is not None:
                self._print(f" AS {node.alias}")
        elif isinstance(node, Join):
            join_type = (
                "LEFT OUTER JOIN" if isinstance(node, LeftOuterJoin)
                else "FULL OUTER JOIN" if isinstance(node, FullOuterJoin)
                else "JOIN"
            )
            self._print(f"{join_type} {self._get_table_name(node.table)}")
            if node.alias is not None:
                self._print(f" AS {node.alias}")
            self._print(" ON ")
            if node.on is None:
                self._print("TRUE")
            else:
                self.print_node(node.on, indent, True)
        elif isinstance(node, Where):
            self._print(f" WHERE {node.expression}")
        elif isinstance(node, RawSource):
            self._print(node.src)

        # Clauses
        elif isinstance(node, CTE):
            self._print("WITH ")
            if node.recursive:
                self._print("RECURSIVE ")
            self._print(f"{self._get_table_name(node.name)} (")
            self._join(node.columns)
            self._print_nl(") AS (")
            for i, s in enumerate(node.selects):
                if i != 0:
                    self._nl()
                    self._indent_print_nl(indent + 2, "UNION ALL")
                self.print_node(s, indent + 1, True)
            self._nl()
            self._print(f") SELECT * FROM {self._get_table_name(node.name)}")
            if not inlined:
                self._print(";")
        # --------------------------------------------------
        # Primitives, Annotations
        # --------------------------------------------------
        elif isinstance(node, (str, int, float, bool, datetime.date, tuple, type(None))):
            self._indent_print(indent, "")
            self._print_value(node)
        # Expressions
        elif isinstance(node, And):
            self._join(node.expr, " AND ", indent)
        elif isinstance(node, Or):
            self._print("( ")
            for i, expr in enumerate(node.expr):
                if i > 0:
                    self._print(" OR ")
                needs_parens = isinstance(expr, And)
                if needs_parens:
                    self._print("( ")
                self.print_node(expr, indent, True)
                if needs_parens:
                    self._print(" )")
            self._print(" )")
        elif isinstance(node, Terminal):
            self._print(node.expr)
        elif isinstance(node, NotNull):
            self._print(f"{node.var} IS NOT NULL")
        elif isinstance(node, NotExists):
            self._print("NOT EXISTS ( ")
            self.print_node(node.expr, indent, True)
            self._print(" )")
        elif isinstance(node, Like):
            self._print(f"{node.left} LIKE {node.right}")
        else:
            raise Exception(f"Missing SQL.pprint({type(node)}): {node}")


import atexit
from datetime import timezone
import datetime
import json
import html
from typing import Union, Sequence, Optional

import logging
from logging import Handler, LogRecord

from opentelemetry import trace
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider, ReadableSpan    
from opentelemetry.sdk.trace.export import BatchSpanProcessor, SpanExporter, SpanExportResult

from relationalai.util.span_tracker import record_is_span, get_span_path_as_str
from relationalai.util.span_tracker import get_span_allowed_attributes_values, get_span_value_from_key
from relationalai.debugging import logger, Span, filter_span_attrs, otel_traceid_to_opentracing, get_current_span

from relationalai.clients.snowflake import Resources

MAX_PAYLOAD_SIZE = 25*1024 # 25KB
_otel_initialized = False


def is_otel_initialized():
    return _otel_initialized

# cache incoming spans up to a limit until we can read them off later
class CachedSpanExporter(SpanExporter):
    def __init__(self, limit:int=500):
        self.limit = limit
        self.spans = []

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        try:
            # discard the earlier spans to keep caching
            if len(self.spans) + len(spans) > self.limit:
                self.spans = self.spans[len(spans):]
            self.spans.extend(spans)
            return SpanExportResult.SUCCESS
        except Exception:
            return SpanExportResult.FAILURE

    def get_spans(self):
        return self.spans


class NativeAppSpanExporter(SpanExporter):
    def __init__(self, resource:Resources, app_name:str):
        self.resource = resource
        self.app_name = app_name
        self._session = None
        atexit.register(self._flush_spans)

    def _flush_spans(self):
        spans = CACHED_SPAN_EXPORTER.get_spans()
        if spans:
            self.export(spans)

    def get_session(self):
        if self._session is None:
            self._session = self.resource.get_sf_session()
        return self._session

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        # loop through generated payloads exporting
        for payload in self._generate_span_payloads(spans):
            self._export_payload(payload)

        return SpanExportResult.SUCCESS

    def _export_payload(self, payload):
        # function is {app_name}.APP.IMPORT_TRACING(object)
        payload = payload.replace("'", "''")
        q = f"CALL {self.app_name}.APP.IMPORT_TRACING('{payload}')"
        try:
            session = self.get_session()
            session.sql(q).collect()
        except Exception as e:
            logging.warning(f"Failed to export spans: {e}")
            pass

    def _generate_span_payloads(self, spans: Sequence[ReadableSpan]):
        remaining_spans = spans

        while remaining_spans:
            payload, remaining_spans = encode_spans_to_otlp_json(remaining_spans, MAX_PAYLOAD_SIZE)
            yield payload

class OtelHandler(Handler):
    """
    This is a logging handler that inserts spans directly into a Snowflake table.

    It uses a queue and a worker thread to buffer spans and then insert them in batches.
    """

    def __init__(self, tracer):
        super().__init__()
        self.span_tracker = OtelSpanTracker()
        Handler.__init__(self)
        self.tracer = tracer

    def emit(self, record: LogRecord):
        if record_is_span(record):
            if isinstance(record.msg, dict):
                try:
                    event = record.msg
                    self.handle_span_event(event)
                except Exception as e:
                    # Prevent logging system crash from bad span record
                    logging.warning(f"Failed to handle span event: {e}")
            else:
                return

    def handle_span_event(self, event: dict):
        try:
            if event["event"] == "span_start":
                self.span_tracker.handle_span_start(event)
            if event["event"] == "span_end":
                span = self.span_tracker.handle_span_end(event)

                if span is not None:
                    trace_id = span.get('trace_id') or 0
                    parent_id = span.get('parent_id') or 0
                    parent_context = trace.set_span_in_context(trace.NonRecordingSpan(trace.SpanContext(
                        trace_id = trace_id,
                        span_id = parent_id,
                        is_remote = False,
                        trace_flags = trace.TraceFlags(0x01)
                    )))

                    with self.tracer.start_as_current_span(
                        span['type'], 
                        context = parent_context,
                        end_on_exit = False,
                        start_time = span['start_timestamp']) as otel_span:
                        
                        for k, v in span['attrs'].items():
                            if isinstance(v, dict):
                                otel_span.set_attribute(k, json.dumps(v))
                            else:
                                otel_span.set_attribute(k, v)

                    otel_span.end(end_time=span['end_timestamp'])
        except Exception as e:
            logging.warning(f"Span handling error: {e}")
            
class OtelSpanTracker:
    def __init__(self):
        self.open_spans = {}

        # pick up already existing spans
        try:
            existing_span = get_current_span()
            while existing_span:
                self.open_spans[str(existing_span.id)] = existing_span
                existing_span = existing_span.parent
        except Exception as e:
            logging.warning(f"Failed to initialize open spans: {e}")

    def handle_span_start(self, msg: dict) -> Union[None, Span]:
        try:
            span: Span = msg["span"]
            span_id = str(span.id)
            self.open_spans[span_id] = span
            return span
        except Exception as e:
            logging.warning(f"Failed to handle span start: {e}")
            return None

    def handle_span_end(self, msg: dict) -> dict:
        try:
            span_id = str(msg["id"])
            span = self.open_spans.pop(span_id)

            joined_key_path = get_span_path_as_str(span)
            end_attrs = get_span_allowed_attributes_values(span, msg["end_attrs"])
            combined_attributes = {
                **span.attrs,  # span_start attrs
                **end_attrs,  # span_end attrs
                "key_path": joined_key_path,
                "elapsed_s": (span.end_timestamp - span.start_timestamp).total_seconds()
            }
            filtered_attrs = filter_span_attrs(combined_attributes)
            formatted_span_msg = format_span_msg(span, filtered_attrs)
            return formatted_span_msg
        except Exception as e:
            logging.warning(f"Failed to handle span end: {e}")
            return {}

#region Span formatting and filtering functions
def format_span_msg(span: Span, attrs: dict) -> dict:
    try:
        span_key = get_span_value_from_key(span)
        attrs["pyrel_span_id"] = span.id_64bit
        trace_id = span.trace_id.int
        attrs["pyrel_trace_id"] = otel_traceid_to_opentracing(trace_id)

        a = {k: encode_attribute(v) for k, v in attrs.items()}

        # ERP uses 64-bit trace IDs (OpenTracing), while OTel uses 128-bit.
        # We downscale to 64-bit for ERP, then pad back to 128-bit for Observe (OTel-compliant).
        return {
            "type": span.type,
            "id": span.id_64bit, 
            "parent_id": span.parent.id_64bit if span.parent else None,
            "trace_id": otel_traceid_to_opentracing(trace_id),
            "key": span_key,
            "start_timestamp": timestamp_to_ns(span.start_timestamp), 
            "end_timestamp": timestamp_to_ns(span.end_timestamp), 
            "attrs": a
        }
    except KeyError as e:
        logging.warning(f"Missing key in span or attrs: {e}")
        return {}
    except AttributeError as e:
        logging.warning(f"Attribute missing in span: {e}")
        return {}
    except Exception as e:
        logging.warning(f"Error formatting span message: {e}")
        return {}


def encode_attribute(attr):
    try:
        if isinstance(attr, datetime.datetime):
            return attr.isoformat() + "Z"
        if isinstance(attr, dict):
            return {k: encode_attribute(v) for k, v in attr.items()}
        if hasattr(attr, "to_dict"):
            return attr.to_dict()
        if hasattr(attr, "to_json"):
            return attr.to_json()
        if isinstance(attr, int) or isinstance(attr, float) or isinstance(attr, bool):
            return attr
        else:
            return str(attr)
    except Exception as e:
        logging.warning(f"Failed to encode attribute {attr}: {e}")
        return None

def timestamp_to_ns(ts):
    try:
        # convert timestamp to UTC and then to ns
        return int(ts.replace(tzinfo=timezone.utc).timestamp() * 1e9)
    except Exception as e:
        logging.warning(f"Failed to convert timestamp {ts} to ns: {e}")
        return 0

def encode_otlp_attribute(k, v):
    try:
        valueObj = {}
        if isinstance(v, str):
            valueObj = {
                "stringValue": html.escape(v).replace("\n", "; ")
            }
        elif isinstance(v, int):
            valueObj = {
                "intValue": v
            }
        elif isinstance(v, float):
            valueObj = {
                "doubleValue": v
            }
        elif isinstance(v, bool):
            valueObj = {
                "boolValue": v
            }
        else:
            logging.warning(f"Unhandled attribute type for key {k}: {type(v)}")
        return {
            "key": k,
            "value": valueObj
        }
    except Exception as e:
        logging.warning(f"Failed to encode attribute {k}: {e}")
        return {"key": k, "value": {"stringValue": "error"}}  # Default error value

def encode_resource_to_otlp_json(resource: Resource) -> str:
    try:
        resource_json = {
            "attributes": []
        }
        
        for k, v in resource.attributes.items():
            try:
                resource_json["attributes"].append(encode_otlp_attribute(k, v))
            except Exception as e:
                logging.warning(f"Failed to encode attribute {k}: {e}")
                resource_json["attributes"].append({"key": k, "value": {"stringValue": "error"}})  # Default error value
        
        return json.dumps(resource_json)
    except Exception as e:
        logging.warning(f"Failed to encode resource to OTLP JSON: {e}")
        return json.dumps({"attributes": []})  # Return empty attributes on failure

def safe_int(value, default=0):
    try:
        return int(value)
    except (ValueError, TypeError):
        return default

def encode_span_to_otlp_json(span: ReadableSpan) -> str:
    try:
        # Take span ID from pyrel, not the one OTelSDK created, to match the parent context
        span_id = safe_int(span.attributes.get("pyrel_span_id")) if span.attributes else 0
        trace_id = safe_int(span.attributes.get("pyrel_trace_id")) if span.attributes else 0

        span_json = {
            "traceId": traceid_to_otlp_json(trace_id),
            "spanId": spanid_to_otlp_json(span_id),
            "parentSpanId": spanid_to_otlp_json(span.parent.span_id) if span.parent else None,
            "name": span.name,
            "startTimeUnixNano": span.start_time,
            "endTimeUnixNano": span.end_time,
            "droppedAttributesCount": 0,
            "events": [],
            "droppedEventsCount": 0,
            "status": {
                "message": span.status.description if span.status else "No status",
                "code": span.status.status_code.value if span.status else 0
            }
        }

        # Add attributes
        span_json["attributes"] = []
        for k, v in (span.attributes or {}).items():
            if k == "pyrel_span_id" or k == "pyrel_trace_id":  # Already encoded in span id
                continue
            try:
                k, v = conform_attributes_to_rai_observability(span.name, k, str(v))
                span_json["attributes"].append(encode_otlp_attribute(k, v))
            except Exception as e:
                logging.warning(f"Error conforming attribute {k}: {e}")
                continue  # Skip this attribute if it can't be conformed

        # Add events
        for event in span.events:
            try:
                event_json = {
                    "timeUnixNano": timestamp_to_ns(event.timestamp),
                    "name": event.name,
                    "droppedAttributesCount": 0
                }
                event_json["attributes"] = []
                for k, v in (event.attributes or {}).items():
                    event_json["attributes"].append(encode_otlp_attribute(k, v))
                span_json["events"].append(event_json)
            except Exception as e:
                logging.warning(f"Error processing event {event.name}: {e}")
                continue  # Skip the event if it can't be processed

        # Return as JSON string
        return json.dumps(span_json)
    except Exception as e:
        logging.warning(f"Error encoding span to OTLP JSON: {e}")
        return json.dumps({})  # Return empty JSON on failure

def encode_spans_to_otlp_json(spans: Sequence[ReadableSpan], max_bytes: int) -> tuple[str, Sequence[ReadableSpan]]:
    if len(spans) == 0:
        return "", []

    try:
        # Get resource from the first span, assume all spans have the same resource
        # which is valid in the PyRel context because there is a single tracer
        resource = spans[0].resource
        resource_str = encode_resource_to_otlp_json(resource)
    except Exception as e:
        logging.warning(f"Error encoding resource to OTLP JSON: {e}")
        return "", spans  # Return the unencoded spans if resource encoding fails
    
    # TODO encode pyrel version for real
    header_str = f'{{"resourceSpans": [{{"resource":{resource_str}, "scopeSpans": [{{"scope":{{"name":"pyrel", "version":"v0.4.0"}},"spans":['
    footer_str = ']}]}]}'
    current_size = len(header_str) + len(footer_str)

    encoded_spans = []
    for span in spans:
        try:
            span_str = encode_span_to_otlp_json(span)
        except Exception as e:
            logging.warning(f"Error encoding span {span}: {e}")
            continue  # Skip this span if encoding fails
        
        # Check if adding this span exceeds the max size
        if current_size + len(span_str) > max_bytes:
            break
        
        encoded_spans.append(span_str)
        current_size += len(span_str)
    return header_str + ",".join(encoded_spans) + footer_str, spans[len(encoded_spans):]

# Format as a 128-bit hex string, padding if necessary
def traceid_to_otlp_json(trace_id: int) -> str:
    try:
        return format(trace_id, 'x').zfill(32)
    except Exception as e:
        logging.warning(f"Error formatting trace_id {trace_id}: {e}")
        return "0" * 32

# Format as a 64-bit hex string
def spanid_to_otlp_json(span_id: int) -> str:
    try:
        return format(span_id, 'x').zfill(16)
    except Exception as e:
        logging.warning(f"Error formatting span_id {span_id}: {e}")
        return "0" * 16  # Return a default value on error

# generate map for any span
PYREL_TO_RAI_KEY_MAP = {
    "txn_id": "rai.transaction_id",
}

# generate map for specific span types
PYREL_SPAN_SPECIFIC_RAI_KEY_MAP = {
    "create_engine": {
        "name": "rai.engine_name",
        "size": "sf.computepool_size",
        "pool": "sf.computepool_name"
    },
    "get_model": {
        "name": "rai.database_name",
    },
    "create_model": {
        "name": "rai.database_name",
    },
    "delete_model": {
        "name": "rai.database_name",
    },
}

def conform_attributes_to_rai_observability(span_type: str, k: str, v: str):
    try:
        k = PYREL_TO_RAI_KEY_MAP.get(k, k)
        if span_type and isinstance(span_type, str):
            k = PYREL_SPAN_SPECIFIC_RAI_KEY_MAP.get(span_type, {}).get(k, k)
    except Exception as e:
        logging.warning(f"Failed to conform attribute ({k}) for span type ({span_type}): {e}")
    return k, v

TRACE_PROVIDER: Optional[TracerProvider] = None

CACHED_SPAN_EXPORTER = CachedSpanExporter()

def disable_otel_handling():
    global TRACE_PROVIDER, _otel_initialized
    try:
        for handler in list(logger.handlers):
            if isinstance(handler, OtelHandler):
                logger.removeHandler(handler)
        TRACE_PROVIDER = None
        _otel_initialized = False
    except Exception as e:
        logging.warning(f"disable_otel_handling failed: {e}")


def enable_otel_export(client_resources: Resources, app_name):
    global TRACE_PROVIDER, _otel_initialized
    if _otel_initialized:
        logging.warning("OTel already initialized, skipping.")
        return
    _otel_initialized = True

    if TRACE_PROVIDER is None:
        TRACE_PROVIDER = TracerProvider(resource=Resource.create({'service.name': 'pyrel'}))
    logger.addHandler(OtelHandler(TRACE_PROVIDER.get_tracer("pyrel")))

    exporter = NativeAppSpanExporter(client_resources, app_name)
    try:
        spans = CACHED_SPAN_EXPORTER.get_spans()
        exporter.export(spans)
    except Exception as e:
        logging.warning(f"Failed to export cached spans: {e}")

    span_processor = BatchSpanProcessor(exporter, 10000) #10s
    TRACE_PROVIDER.add_span_processor(span_processor)
    



from typing import Sequence, Optional

import pandas as pd
import numpy as np
import datetime
import pyarrow as pa
from snowflake.connector.cursor import ResultMetadata
from snowflake.connector.constants import FIELD_ID_TO_NAME

from relationalai.early_access.metamodel import ir
from relationalai.clients.result_helpers import Int128Dtype
from relationalai.early_access.metamodel.types import Int64, Int128, DateTime


def format_columns(result_frame:pd.DataFrame, result_metadata:Sequence[ResultMetadata],
                   original_columns_metadata: dict[str, Optional[ir.Type]]) -> pd.DataFrame:
    for i, col in enumerate(result_frame.columns):
        col_name = col.lower()
        col_metadata = result_metadata[i]
        col_type = original_columns_metadata[col_name] if col_name in original_columns_metadata else None
        if col_type == DateTime:
            result_frame[col] = result_frame[col].dt.floor("ms")
        elif col_type and isinstance(col_type, ir.DecimalType):
            pa_dtype = pa.decimal128(precision=col_type.precision, scale=col_type.scale)
            result_frame[col] = result_frame[col].astype(pd.ArrowDtype(pa_dtype))
        elif result_frame[col].apply(lambda x: isinstance(x, (datetime.date, datetime.datetime))).any():
            result_frame[col] = pd.to_datetime(result_frame[col], errors='coerce')
        elif FIELD_ID_TO_NAME[col_metadata.type_code] == "FIXED" and col_metadata.scale == 0:
            series = result_frame[col]

            # Prefer explicit type from original metadata if present
            if col_type == Int64:
                result_frame[col] = series.astype("Int64")
            elif col_type == Int128:
                result_frame[col] = series.astype(Int128Dtype())
            elif col_metadata.precision:
                result_frame[col] = _cast_integer_column(series, col_metadata.precision)


        # SQL may return None for nulls; replace with np.nan unless the column is FIXED, because
        # both Int64 and Int128 already deal with nulls using pd.NA
        if not FIELD_ID_TO_NAME[col_metadata.type_code] == "FIXED":
            result_frame[col] = result_frame[col].replace({None: np.nan})
    return result_frame

def format_duckdb_columns(result_frame: pd.DataFrame, arrow_schema: pa.Schema) -> pd.DataFrame:
    for field in arrow_schema:
        col = field.name
        pa_type = field.type
        if pa.types.is_decimal(pa_type):
            if pa_type.scale == 0:  # integer-like
                result_frame[col] = _cast_integer_column(result_frame[col], pa_type.precision)
    return result_frame

def _cast_integer_column(series: pd.Series, precision: int) -> pd.Series:
    """Cast a numeric pandas Series to Int64 or Int128 depending on precision."""
    return series.astype("Int64") if precision <= 19 else series.astype(Int128Dtype())



__version__ = "0.0.0"

# NOTE(coey) using Solver class from old solvers library to avoid code duplication
from relationalai.experimental.solvers import Solver, Provider

from .common import make_name, all_different, implies
from .solvers_pb import SolverModelPB
from .solvers_dev import SolverModelDev

def SolverModel(model, num_type: str, use_pb: bool = True):
    if use_pb:
        return SolverModelPB(model, num_type)
    else:
        return SolverModelDev(model, num_type)

__all__ = [
    "Solver",
    "Provider",
    "SolverModel",
    "make_name",
    "all_different",
    "implies",
]



import json
import re
import textwrap
import uuid
from collections import defaultdict
from typing import Any, List, Optional

from pandas import DataFrame
import relationalai as rai
from relationalai import debugging
from relationalai.clients import result_helpers
from relationalai.clients.snowflake import APP_NAME
from relationalai.early_access.dsl.ir.compiler import Compiler
from relationalai.early_access.dsl.ontologies.models import Model
from relationalai.early_access.metamodel import ir
from relationalai.early_access.metamodel.visitor import collect_by_type
from relationalai.early_access.rel.compiler import ModelToRel
from relationalai.clients.config import Config
from relationalai.tools.constants import USE_GRAPH_INDEX, Generation


class RelExecutor:

    def __init__(self, database: str, dry_run: bool = False, config: Optional[Config] = None) -> None:
        super().__init__()
        self.database = database
        self.dry_run = dry_run
        self.compiler = Compiler()
        self.model_to_rel = ModelToRel()
        self.config = config or Config()
        self._resources = None
        self._last_model = None
        self._last_raw_sources = None

    @property
    def resources(self):
        if not self._resources:
            with debugging.span("create_session"):
                self.dry_run |= bool(self.config.get("compiler.dry_run", False))
                self._resources = rai.clients.snowflake.Resources(
                    dry_run=self.dry_run,
                    config=self.config,
                    generation=Generation.QB,
                )
                if not self.dry_run:
                    try:
                        if not self._resources.get_database(self.database) and not self._use_graph_index():
                            self._resources.create_graph(self.database)
                    except Exception as e:
                        if "already exists" not in str(e).lower():
                            raise e
                    self.engine = self._resources.config.get("engine", strict=False)
                    if not self.engine:
                        self.engine = self._resources.get_default_engine_name()
        return self._resources

    def report_errors(self, problems: list[dict[str, Any]], abort_on_error=True):
        from relationalai import errors
        all_errors = []
        undefineds = []
        pyrel_errors = defaultdict(list)
        pyrel_warnings = defaultdict(list)

        for problem in problems:
            message = problem.get("message", "")
            report = problem.get("report", "")
            # TODO: we need to build source maps
            # path = problem.get("path", "")
            # source_task = self._install_batch.line_to_task(path, problem["start_line"]) or task
            # source = debugging.get_source(source_task) or debugging.SourceInfo()
            source = debugging.SourceInfo()
            severity = problem.get("severity", "warning")
            code = problem.get("code")

            if severity in ["error", "exception"]:
                if code == "UNDEFINED_IDENTIFIER":
                    match = re.search(r'`(.+?)` is undefined', message)
                    if match:
                        undefineds.append((match.group(1), source))
                    else:
                        all_errors.append(errors.RelQueryError(problem, source))
                elif "overflowed" in report:
                    all_errors.append(errors.NumericOverflow(problem, source))
                elif code == "PYREL_ERROR":
                    pyrel_errors[problem["props"]["pyrel_id"]].append(problem)
                elif abort_on_error:
                    all_errors.append(errors.RelQueryError(problem, source))
            else:
                if code == "ARITY_MISMATCH":
                    errors.ArityMismatch(problem, source)
                elif code == "IC_VIOLATION":
                    all_errors.append(errors.IntegrityConstraintViolation(problem, source))
                elif code == "PYREL_ERROR":
                    pyrel_warnings[problem["props"]["pyrel_id"]].append(problem)
                else:
                    errors.RelQueryWarning(problem, source)

        if abort_on_error and len(undefineds):
            all_errors.append(errors.UninitializedPropertyException(undefineds))

        if abort_on_error:
            for pyrel_id, pyrel_problems in pyrel_errors.items():
                all_errors.append(errors.ModelError(pyrel_problems))

        for pyrel_id, pyrel_problems in pyrel_warnings.items():
            errors.ModelWarning(pyrel_problems)

        if len(all_errors) == 1:
            raise all_errors[0]
        elif len(all_errors) > 1:
            raise errors.RAIExceptionSet(all_errors)

    def _use_graph_index(self):
        return self.config.get("use_graph_index", USE_GRAPH_INDEX)

    def _export(self, raw_code: str, dest_table: str, actual_cols: list[str], declared_cols: list[str], update:bool):
        _exec = self.resources._exec
        output_table = "out" + str(uuid.uuid4()).replace("-", "_")
        txn_id = None
        artifacts = None
        assert self.resources._session
        with debugging.span("transaction"):
            try:
                with debugging.span("exec_format") as span:
                    if self._use_graph_index():
                        res = _exec(f"call {APP_NAME}.api.exec_into_table(?, ?, ?, ?, ?, ?);",
                                         [self.database, self.engine, raw_code, output_table, False, True])
                        txn_id = json.loads(res[0]["EXEC_INTO_TABLE"])["rai_transaction_id"]
                    else:
                        res = _exec(f"call {APP_NAME}.api.exec_into(?, ?, ?, ?, ?, {{}}, ?, {{}});",
                                    [self.database, self.engine, raw_code, output_table, False, True])
                        txn_id = json.loads(res[0]["EXEC_INTO"])["rai_transaction_id"]
                    span["txn_id"] = txn_id

                with debugging.span("write_table", txn_id=txn_id):
                    out_sample = _exec(f"select * from {APP_NAME}.results.{output_table} limit 1;")
                    if out_sample:
                        keys = set([k.lower() for k in out_sample[0].as_dict().keys()])
                        fields = []
                        ix = 0
                        for name in declared_cols:
                            if name in actual_cols:
                                field = f"col{ix:03} as {name}" if f"col{ix:03}" in keys else f"NULL as {name}"
                                ix += 1
                            else:
                                field = f"NULL as {name}"
                            fields.append(field)
                        names = ", ".join(fields)
                        if not update:
                            _exec(f"""
                                begin
                                    CREATE OR REPLACE TABLE {dest_table} AS SELECT {names} FROM {APP_NAME}.results.{output_table};
                                    call {APP_NAME}.api.drop_result_table('{output_table}');
                                end;
                            """)
                        else:
                            _exec(f"""
                                begin
                                    INSERT INTO {dest_table} SELECT {names} FROM {APP_NAME}.results.{output_table};
                                    call {APP_NAME}.api.drop_result_table('{output_table}');
                                end;
                            """)

            except Exception as e:
                msg = str(e).lower()
                if "no columns returned" in msg or "columns of results could not be determined" in msg:
                    pass
                else:
                    raise e
            if txn_id:
                artifact_info = self.resources._list_exec_async_artifacts(txn_id)
                with debugging.span("fetch"):
                    artifacts = self.resources._download_results(artifact_info, txn_id, "ABORTED")
            return artifacts

    def execute_model(self, model: Model, result_cols: Optional[List[str]] = None) -> DataFrame:
        raw_sources_ir = self.compiler.compile_raw_sources(model.raw_sources()) if model.raw_sources() else None
        ir_model = self.compiler.compile_model(model)
        exports = model.exports()
        queries = model.queries()
        if exports:
            for export in exports:
                if export.target_table is None:
                    raise ValueError("Export target table must be specified")
                export_ir_model = self.compiler.compile_export(export)
                self.execute(ir_model, export_ir_model, raw_sources_ir, export.labels, export.target_table)
            if not queries:
                return DataFrame()
        query_ir_model = self.compiler.compile_queries(model.queries())
        return self.execute(ir_model, query_ir_model, raw_sources_ir, result_cols)

    def execute(self,
                model: ir.Model,
                task: ir.Model,
                raw_sources: Optional[ir.Model] = None,
                result_cols: Optional[List[str]] = None,
                export_to:Optional[str]=None,
                update:bool=False,
                ) -> DataFrame:

        resources = self.resources

        if raw_sources and self._last_raw_sources != raw_sources:
            with debugging.span("compile", metamodel=task) as compile_span:
                raw_sources_code = str(self.model_to_rel.to_rel(raw_sources, options={"no_declares": True}))
                compile_span["compile_type"] = "raw_sources"
                compile_span["rel"] = raw_sources_code
                if not self.dry_run:
                    code = textwrap.dedent(f"""{raw_sources_code}""")
                    raw_results = self.resources.exec_raw(self.database, self.engine, code, False, nowait_durable=True)
                    _, errs = result_helpers.format_results(raw_results, None, result_cols)  # Pass None for task parameter
                    self.report_errors(errs)
                    if not errs:
                        self._last_raw_sources = raw_sources

        rules_code = ""
        if self._last_model != model:
            with debugging.span("compile", metamodel=model) as install_span:
                rules_code = str(self.model_to_rel.to_rel(model))
                install_span["compile_type"] = "model"
                install_span["rel"] = rules_code
                rules_code = resources.create_models_code([("pyrel_qb_0", rules_code)])
                self._last_model = model

        with debugging.span("compile", metamodel=task) as compile_span:
            task_code = str(self.model_to_rel.to_rel(task, options={"no_declares": True}))
            compile_span["compile_type"] = "task"
            compile_span["rel"] = task_code

        full_code = textwrap.dedent(f"""
            {rules_code}
            {task_code}
        """)

        if self.dry_run:
            return DataFrame()

        outputs = collect_by_type(ir.Output, task)
        cols = None
        if outputs:
            cols = [alias for alias, _ in outputs[-1].aliases if alias]

        if not export_to:
            raw_results = self.resources.exec_raw(self.database, self.engine, full_code, False, nowait_durable=True)
            df, errs = result_helpers.format_results(raw_results, None, result_cols)  # Pass None for task parameter
            self.report_errors(errs)
            return df
        else:
            assert cols
            # The result cols should be a superset of the actual cols.
            if result_cols is not None:
                assert all(col in result_cols for col in cols)
            else:
                result_cols = cols
            assert result_cols
            raw = self._export(full_code, export_to, cols, result_cols, update)
            if raw:
                dataframe, errors = result_helpers.format_results(raw, None, result_cols)
                self.report_errors(errors)
            return DataFrame()


from __future__ import annotations
from typing import Any, Iterable, Sequence as PySequence, cast, Tuple, Union
from dataclasses import dataclass, field
from decimal import Decimal as PyDecimal

from relationalai.early_access.metamodel import ir, compiler as c, builtins as bt, types, visitor, helpers, factory as f
from relationalai.early_access.metamodel.typer import Checker, InferTypes
from relationalai.early_access.metamodel.typer.typer import to_base_primitive, to_type, _NON_PARAMETRIC_PRIMITIVES
from relationalai.early_access.metamodel.rewrite import Flatten
from relationalai.early_access.metamodel.rewrite import ExtractKeys

from relationalai.early_access.metamodel.rewrite import Splinter
from relationalai.early_access.metamodel.rewrite import ExtractNestedLogicals
from relationalai.early_access.metamodel.visitor import ReadWriteVisitor
from relationalai.early_access.metamodel.util import OrderedSet, group_by, NameCache, ordered_set

from relationalai.early_access.rel import rel, rel_utils as u, rewrite, builtins as rel_bt

import math


#--------------------------------------------------
# Compiler
#--------------------------------------------------

class Compiler(c.Compiler):
    def __init__(self):
        super().__init__([
            Checker(),
            ExtractNestedLogicals(), # before InferTypes to avoid extracting casts
            InferTypes(),
            rewrite.CDC(),
            ExtractKeys(),
            # rewrite.ExtractCommon(),
            Flatten(),
            Splinter(),
            rewrite.QuantifyVars(),
        ])
        self.model_to_rel = ModelToRel()

    def do_compile(self, model: ir.Model, options:dict={}) -> str:
        return str(self.model_to_rel.to_rel(model, options=options))

COMPILER_OPTIONS = [
    # do not generated declarations for relations read by the model but not written to
    "no_declares",
    # do not GNF the output relation, keeping it wide
    "wide_outputs"
]

@dataclass
class ModelToRel:
    """ Generates Rel from an IR Model, assuming the compiler rewrites were done. """

    relation_name_cache: NameCache = field(default_factory=NameCache)
    rule_name_cache: NameCache = field(default_factory=NameCache)

    # Map a rel variable to one with a different name
    var_map: dict[rel.Var, rel.Var] = field(default_factory=dict)

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------

    def to_rel(self, model: ir.Model, options:dict = {}) -> rel.Program:
        self._register_external_relations(model)

        rules = self._generate_rules(model)
        reads = self._rel_reads(rules)
        declares = [] if options.get("no_declares") else self._generate_declares(model)
        self._rel_reads(declares, reads)
        return rel.Program(tuple([
            *self._generate_builtin_defs(model, reads),
            *declares,
            *rules,
        ]))

    #--------------------------------------------------
    # Top level handlers
    #--------------------------------------------------

    def _generate_builtin_defs(self, model: ir.Model, reads:OrderedSet[str]) -> list[rel.Def]:
        defs = []

        for t in model.types:
            # TODO - some of these types are never used, we should not generate them, but we
            # need to replace the rel_reads function with something that collects those types
            # during generation
            if isinstance(t, ir.DecimalType):
                defs.append(
                    rel.Def(u.rel_typename(t),
                        tuple([rel.Var("x")]),
                        rel.atom("::std::common::FixedDecimal",
                            tuple([rel.MetaValue(types.digits_to_bits(t.precision)), rel.MetaValue(t.scale), rel.Var("x")])),
                        tuple([rel.Annotation("inline", ())]),
                    ),
                )

        if "pyrel_inf" in reads:
            defs.append(
                rel.Def("pyrel_inf",
                    tuple([rel.Var("x")]),
                    rel.atom("::std::common::infinity", tuple([rel.MetaValue(64), rel.Var("x")])),
                    tuple([rel.Annotation("inline", ())]),
                ),
            )

        if "pyrel_NaN" in reads:
            defs.append(
                rel.Def("pyrel_NaN",
                    tuple([rel.Var("x")]),
                    rel.atom("::std::common::nan", tuple([rel.MetaValue(64), rel.Var("x")])),
                    tuple([rel.Annotation("inline", ())]),
                ),
            )

        if "pyrel_ID" in reads:
            defs.append(rel.Def("pyrel_ID",
                tuple([rel.Var("x")]),
                rel.Or(OrderedSet[rel.Expr].from_iterable([
                    rel.atom("::std::common::UInt128", tuple([rel.Var("x")])),
                    rel.atom("::std::common::Missing", tuple([rel.Var("x")])),
                ])),
                tuple([rel.Annotation("inline", ())]),
            ))

        if "pyrel_num_chars" in reads:
            defs.append(
                rel.Def("pyrel_num_chars",
                    tuple([rel.Var("x"), rel.Var("y")]),
                    rel.Exists(
                        tuple([rel.Var("z")]),
                        rel.And(ordered_set(
                            rel.atom("::std::common::num_chars", tuple([rel.Var("x"), rel.Var("z")])),
                            rel.Atom(self._convert_abs(types.Int64, types.Int128), tuple([rel.Var("z"), rel.Var("y")])),
                        )),
                    ),
                    tuple([rel.Annotation("inline", ())]),
                ),
            )

        if "pyrel_count" in reads:
            defs.append(
                rel.Def("pyrel_count",
                    tuple([rel.Var("{R}"), rel.Var("y")]),
                    rel.Exists(
                        tuple([rel.Var("z")]),
                        rel.And(ordered_set(
                            rel.atom("::std::common::count", tuple([rel.Var("R"), rel.Var("z")])),
                            rel.Atom(self._convert_abs(types.Int64, types.Int128), tuple([rel.Var("z"), rel.Var("y")])),
                        )),
                    ),
                    tuple([rel.Annotation("inline", ())]),
                ),
            )

        if "pyrel_sort" in reads:
            defs.append(
                rel.Def("pyrel_sort",
                    tuple([rel.Var("{R}"), rel.Var("n"), rel.Var("r", varargs=True)]),
                    rel.Exists(
                        tuple([rel.Var("z")]),
                        rel.And(ordered_set(
                            rel.atom("::std::common::sort", tuple([rel.Var("R"), rel.Var("z"), rel.Var("r", varargs=True)])),
                            rel.Atom(self._convert_abs(types.Int64, types.Int128), tuple([rel.Var("z"), rel.Var("n")])),
                        )),
                    ),
                    tuple([rel.Annotation("inline", ())]),
                ),
            )

        if "pyrel_reverse_sort" in reads:
            defs.append(
                rel.Def("pyrel_reverse_sort",
                    tuple([rel.Var("{R}"), rel.Var("n"), rel.Var("r", varargs=True)]),
                    rel.Exists(
                        tuple([rel.Var("z")]),
                        rel.And(ordered_set(
                            rel.atom("::std::common::reverse_sort", tuple([rel.Var("R"), rel.Var("z"), rel.Var("r", varargs=True)])),
                            rel.Atom(self._convert_abs(types.Int64, types.Int128), tuple([rel.Var("z"), rel.Var("n")])),
                        )),
                    ),
                    tuple([rel.Annotation("inline", ())]),
                ),
            )

        if "pyrel_top" in reads:
            defs.append(
                rel.Def("pyrel_top",
                    tuple([rel.Var("k"), rel.Var("{R}"), rel.Var("n"), rel.Var("r", varargs=True)]),
                    rel.Exists(
                        tuple([rel.Var("z")]),
                        rel.And(ordered_set(
                            rel.atom("::std::common::top", tuple([rel.Var("k"), rel.Var("R"), rel.Var("z"), rel.Var("r", varargs=True)])),
                            rel.Atom(self._convert_abs(types.Int64, types.Int128), tuple([rel.Var("z"), rel.Var("n")])),
                        )),
                    ),
                    tuple([rel.Annotation("inline", ())]),
                ),
            )

        if "pyrel_bottom" in reads:
            defs.append(
                rel.Def("pyrel_bottom",
                    tuple([rel.Var("k"), rel.Var("{R}"), rel.Var("n"), rel.Var("r", varargs=True)]),
                    rel.Exists(
                        tuple([rel.Var("z")]),
                        rel.And(ordered_set(
                            rel.atom("::std::common::bottom", tuple([rel.Var("k"),rel.Var("R"), rel.Var("z"), rel.Var("r", varargs=True)])),
                            rel.Atom(self._convert_abs(types.Int64, types.Int128), tuple([rel.Var("z"), rel.Var("n")])),
                        )),
                    ),
                    tuple([rel.Annotation("inline", ())]),
                ),
            )

        if "pyrel_dates_period_days" in reads:
            defs.append(
                rel.Def("pyrel_dates_period_days",
                    tuple([rel.Var("a"), rel.Var("b"), rel.Var("c")]),
                    rel.Exists(
                        tuple([rel.Var("d")]),
                        rel.And(ordered_set(
                            rel.atom("::std::common::dates_period_days",
                                     tuple([rel.Var("a"), rel.Var("b"), rel.Var("d")])),
                            rel.atom("::std::common::^Day",
                                     tuple([rel.Var("c"), rel.Var("d")]))
                        )),
                    ),
                    tuple([rel.Annotation("inline", ())]),
                    ),
            )

        if "pyrel_datetimes_period_milliseconds" in reads:
            defs.append(
                rel.Def("pyrel_datetimes_period_milliseconds",
                    tuple([rel.Var("a"), rel.Var("b"), rel.Var("c")]),
                    rel.Exists(
                        tuple([rel.Var("m")]),
                        rel.And(ordered_set(
                            rel.atom("::std::common::datetimes_period_milliseconds",
                                     tuple([rel.Var("a"), rel.Var("b"), rel.Var("m")])),
                            rel.atom("::std::common::^Millisecond",
                                     tuple([rel.Var("c"), rel.Var("m")]))
                        )),
                    ),
                    tuple([rel.Annotation("inline", ())]),
                    ),
            )

        return defs

    @staticmethod
    def _convert_abs(from_type: ir.Type, to_type: ir.Type):
        if from_type == types.Int64 and to_type == types.Float:
            return rel.Identifier("::std::common::int_float_convert")
        elif from_type == types.Float and to_type == types.Int64:
            return rel.Identifier("::std::common::float_int_convert")
        else:
            input_type = u.rel_typename(from_type)
            output_type = u.rel_typename(to_type)
            return rel.RelationalAbstraction(
                tuple([rel.Var("val_x", type=input_type), rel.Var("val_y", type=output_type)]),
                rel.Exists(
                    tuple([rel.Var("type_x"), rel.Var("type_y")]),
                    rel.And(ordered_set(
                        # Since we declared them to be the types we're converting from and to, we can just use the types of x and y here.
                        # The Rel compiler will use the static type of the variable to compute the Type values.
                        rel.atom("rel_primitive_typeof", tuple([rel.Var("val_x"), rel.Var("type_x")])),
                        rel.atom("rel_primitive_typeof", tuple([rel.Var("val_y"), rel.Var("type_y")])),
                        rel.atom("rel_primitive_convert", tuple([rel.Var("type_x"), rel.Var("type_y"), rel.Var("val_x"), rel.Var("val_y")])),
                    )),
                )
            )

    def _generate_declares(self, m: ir.Model) -> list[rel.Declare]:
        """
        Generate declare statements for relations declared by the model and:
            - not built-ins
            - not used as an annotation
            - not annotated as external
            - do not start with ^ (for hardcoded Rel constructors)
            - and are never the target of an update
        """
        rw = ReadWriteVisitor()
        m.accept(rw)

        root = cast(ir.Logical, m.root)

        annotations = [anno.relation for anno in visitor.collect_by_type(ir.Annotation, m.root)]
        reads = m.relations - rw.writes(root) - bt.builtin_relations - bt.builtin_overloads - bt.builtin_annotations - annotations
        reads = list(filter(lambda r: not r.name.startswith('^') and not helpers.is_external(r), reads))

        primitive_type_names = OrderedSet.from_iterable([t.name for t in _NON_PARAMETRIC_PRIMITIVES])
        declares: list[rel.Declare] = []
        for r in reads:
            if r.name in rel.infix or r.name in u.OPERATORS:
                continue

            if helpers.is_from_cast(r):
                continue

            # TODO: should address the root of the issue
            # In some cases we might end up with explicit Concepts for primitives in the model
            if r.name in primitive_type_names or r.name.startswith("Decimal("):
                continue

            # In case parameter name starts with ':' use its name instead of type name
            def requires_name(fld: ir.Field):
                if isinstance(fld.type, ir.ScalarType):
                    t = fld.type
                    if t == types.Symbol:
                        return rel.MetaValue(fld.name[1:])
                    else:
                        return rel.Var(name=u.sanitize(fld.name.lower()), type=u.rel_typename(t))
                else:
                    return rel.Var(u.sanitize(fld.name.lower()))
            head = tuple([requires_name(f) for f in r.fields])

            # Example: declare test(:a, _x0 in Int, _x1 in String) requires true
            declares.append(rel.Declare(
                rel.atom(self._relation_name(r), head),
                rel.true  # `requires true` does not generate any constraints, that affects performance on the RAI side
            ))
        return declares

    def _generate_rules(self, m: ir.Model) -> list[Union[rel.Def, rel.RawSource]]:
        """ Generate rules for the root of this model.

        Assumes the model already was processed such that it contains a root Logical with
        children that are also Logical tasks representing the rules to generate.
        """
        rules: list[Union[rel.Def, rel.RawSource]] = []
        root = cast(ir.Logical, m.root)
        for child in root.body:
            rules.extend(self._generate_rule(cast(ir.Logical, child)))
        return rules

    def _generate_rule(self, rule: ir.Logical) -> list[Union[rel.Def, rel.RawSource]]:
        """ Generate rules for a nested Logical in a model.

        This is for a top-level Logical, under the root Logical.
        """
        # reset the name cache for each rule
        self.rule_name_cache = NameCache()
        effects, other, aggregates, ranks = self._split_tasks(rule.body)
        if not effects or (aggregates and ranks):
            # nothing to generate for this Logical
            return []

        elif len(effects) == 1:
            # a single effect with a body becomes a single rule
            effect = effects[0]

            # deal with raw sources
            if isinstance(effect, ir.Update) and effect.relation == bt.raw_source:
                # TODO: remove this once the type checker checks this.
                assert(len(effect.args) == 2 and isinstance(effect.args[0], ir.Literal) and isinstance(effect.args[1], ir.Literal))
                if effect.args[0].value != "rel":
                    return []
                return [rel.RawSource(cast(str, effect.args[1].value))]
            else:
                args, lookups, rel_equiv = self._effect_args(effect)
                if lookups:
                    other.extend(lookups)
                return [rel.Def(
                    self._effect_name(effect),
                    args,
                    rel.create_and([
                        self.generate_logical_body(other, aggregates, ranks),
                        *rel_equiv
                    ]),
                    self.generate_annotations(effect.annotations)
                )]
        else:
            # currently we can only deal with multiple effects if they are all updates with
            # no body, which is the pattern for inserting hardcoded data.
            if other or aggregates:
                raise NotImplementedError("Body in logical task with multiple effects.")
            if any(isinstance(effect, ir.Output) for effect in effects):
                raise NotImplementedError("Output in logical task with multiple effects.")
            sample = cast(ir.Update, effects[0]).effect
            if any(cast(ir.Update, effect).effect != sample for effect in effects):
                raise NotImplementedError("Different types of effects in logical task.")

            # Group updates by relation name
            relation_groups = group_by(cast(list[ir.Update], effects), lambda e: self._relation_name(e.relation))

            # Process each relation group
            defs = []
            for name, updates in relation_groups.items():
                effects_to_union = []
                for update in updates:
                    update_args, lookups, rel_equiv = self._effect_args(update)
                    if update_args:
                        defs.append(
                            rel.Def(
                                name,
                                update_args,
                                rel.create_and([
                                    self.generate_body_expr(lookups),
                                    *rel_equiv
                                ]),
                                self.generate_annotations(update.annotations)
                            )
                        )
                    else:
                        effects_to_union.append(update)

                if effects_to_union:
                    update = updates.some()
                    args, lookups, rel_equiv = self._effect_args(update)
                    bodies = []
                    for update in effects_to_union:
                        bodies.append(self.handle(update))
                    for lookup in lookups:
                        bodies.append(self.handle(lookup))

                    defs.append(
                        rel.Def(
                            name,
                            args,
                            rel.create_and([
                                rel.Union(tuple(bodies)),
                                *rel_equiv
                            ]),
                            self.generate_annotations(update.annotations)
                        )
                    )
            return defs

    def generate_logical_body(self, other, aggregates, ranks):
        """ Generate the body of a rule for a Logical that contains these aggregates/ranks
        and other tasks (i.e., no effects)."""

        if aggregates:
            # push the body into the aggregates; this assumes a rewrite pass already
            # prepared the body to contain only what's needed by the aggregates
            exprs = []
            for agg in aggregates:
                # The variables declared in the relational abstraction are the agg's "projection" + "over"
                abs_vars = OrderedSet.from_iterable(agg.projection)
                result = []
                for arg in agg.args:
                    if helpers.is_aggregate_input(arg, agg):
                        new_arg = arg if isinstance(arg, ir.Var) else self.handle(arg)
                        abs_vars.add(new_arg)
                    else:
                        result.append(self.handle(arg))

                old_var_map = self.var_map
                self.var_map = {}

                common_vars = OrderedSet.from_iterable(agg.projection) & agg.group
                abs_body_exprs = []
                for v in common_vars:
                    orig_rel_v = self.handle_var(v)
                    inner_rel_v = rel.Var("_t" + orig_rel_v.name)
                    self.var_map[orig_rel_v] = inner_rel_v
                    eq_expr = rel.BinaryExpr(orig_rel_v, "=", inner_rel_v)
                    abs_body_exprs.append(eq_expr)

                abs_head = self.handle_list(tuple(abs_vars))
                abs_body = self.generate_body_expr(other)
                if abs_body_exprs:
                    abs_body = rel.create_and([abs_body, *abs_body_exprs])
                rel_abstraction = rel.RelationalAbstraction(abs_head, abs_body)

                self.var_map = old_var_map

                exprs.append(rel.atom(
                    u.rel_operator(agg.aggregation.name),
                    tuple([ rel_abstraction, *result ])
                ))
            return exprs[0] if len(exprs) == 1 else rel.create_and(exprs)
        elif ranks:
            # push the body into the aggregates; this assumes a rewrite pass already
            # prepared the body to contain only what's needed by the aggregates
            exprs = []
            for rank in ranks:
                rel_name, has_limit = self.compute_rank_limit_info(rank)
                old_var_map = self.var_map
                self.var_map = {}

                abs_vars = ordered_set()
                abs_body_exprs = []
                # Rename the sorted vars to avoid conflicts with the result vars.
                # We sort the requested args, augmented with the keys (projection).
                # The keys have to be present to preserve bag semantics, but should
                # not affect the ranking. Thus they have to go at the end of the list.
                # Create a set to deduplicate vars appearing in both.
                raw_args = OrderedSet.from_iterable(rank.args + rank.projection)
                for ir_v in raw_args:
                    orig_rel_v = self.handle_var(ir_v)
                    if ir_v in rank.projection and ir_v not in rank.group:
                        inner_rel_v = rel.Var("_t" + orig_rel_v.name)
                        self.var_map[orig_rel_v] = inner_rel_v
                    else:
                        inner_rel_v = rel.Var("_t" + orig_rel_v.name)
                        self.var_map[orig_rel_v] = inner_rel_v
                        # inner_rel_v = rel.Var(orig_rel_v.name)
                    if ir_v in rank.group:
                        eq_expr = rel.BinaryExpr(orig_rel_v, "=", inner_rel_v)
                        abs_body_exprs.append(eq_expr)
                    abs_vars.add(inner_rel_v)

                abs_body = self.generate_body_expr(other)
                if abs_body_exprs:
                    abs_body = rel.create_and([abs_body, *abs_body_exprs])
                rel_abstraction = rel.RelationalAbstraction(tuple(abs_vars), abs_body)

                self.var_map = old_var_map

                out_vars = [self.handle_var(v) for v in raw_args]
                params = [rel_abstraction, self.handle_var(rank.result), *out_vars]
                if has_limit:
                    params.insert(0, rank.limit)
                exprs.append(rel.atom(rel_name, tuple(params)))

            return exprs[0] if len(exprs) == 1 else rel.create_and(exprs)
        else:
            # no aggregates or ranks, just return an expression for the body
            return self.generate_body_expr(other)

    def compute_rank_limit_info(self, rank: ir.Rank):
        if all(o for o in rank.arg_is_ascending):
            ascending = True
        elif all(not o for o in rank.arg_is_ascending):
            ascending = False
        else:
            raise Exception("Mixed orderings in rank are not supported yet.")
        has_limit = rank.limit != 0

        if ascending:
            rel_name = "pyrel_top" if has_limit else "pyrel_sort"
        else:
            rel_name = "pyrel_bottom" if has_limit else "pyrel_reverse_sort"
        return rel_name, has_limit

    def generate_body_expr(self, tasks: list[ir.Task]):
        """ Helper to generate the an expression from the tasks, wrapping in Ands if necessary. """
        if not tasks:
            return rel.true
        elif len(tasks) == 1:
            return self.handle(tasks[0])
        else:
            return rel.create_and([self.handle(b) for b in tasks])

    #--------------------------------------------------
    # IR handlers
    #--------------------------------------------------

    def handle(self, n: ir.Node):
        """ Dispatch to the appropriate ir.Node handler. """
        handler = getattr(self, f"handle_{n.kind}", None)
        if handler:
            return handler(n)
        else:
            raise Exception(f"Rel Compiler handler for '{n.kind}' node not implemented.")

    def handle_list(self, n: Iterable[ir.Node]):
        """ Dispatch each node to the appropriate ir.Node handler. """
        return tuple([self.handle(x) for x in n])

    def handle_value(self, type: ir.Type|None, value: Any) -> Union[rel.Primitive, rel.RelationalAbstraction, rel.MetaValue, rel.Var]:
        """ Handle the value (Node or Value) and wrap in a Metavalue if the type is Symbol. """
        # only handle if it is a Node (e.g. ir.Var or ir.Literal)
        v = self.handle(value) if isinstance(value, ir.Node) else value

        # type might be None for these so we have to handle them before the check below.
        if isinstance(v, float) and (math.isinf(v) or math.isnan(v)):
            x = rel.Var("_float")
            rel_name = "::std::common::infinity" if math.isinf(v) else "::std::common::nan"
            return rel.RelationalAbstraction(
                tuple([x]),
                rel.atom(rel_name, tuple([rel.MetaValue(64), x])),
            )

        if type is None:
            return cast(Union[rel.Primitive, rel.RelationalAbstraction, rel.MetaValue, rel.Var], v)

        # only wrap if v is a primitive (i.e. not a metavalue or a var, for example).
        base = to_base_primitive(type) or type
        if type == types.Symbol and isinstance(v, (str, int, float, bool)):
            return rel.MetaValue(v)
        elif isinstance(v, PyDecimal):
            if base == types.GenericDecimal:
                # generic decimals arrive here if that's the type of the field in the relation;
                # in that case, the typer made sure that the value is a Literal and the type in
                # the literal is the exact decimal we need. We can clean this up once we remove
                # support for native literals here, since it's not supported in the IR anymore.
                assert isinstance(value, ir.Literal)
                type = value.type
                assert isinstance(type, ir.DecimalType)
                precision = types.digits_to_bits(type.precision)
            else:
                # if it's not an GenericDecimal, it is a specific decimal and we just use it.
                assert isinstance(type, ir.DecimalType)
                precision = types.digits_to_bits(type.precision)
            x = rel.Var("_dec")
            return rel.RelationalAbstraction(
                tuple([x]),
                rel.atom("::std::common::decimal", tuple([rel.MetaValue(precision), rel.MetaValue(type.scale), v, x]))
            )

        elif base == types.Int128 and isinstance(v, int):
            x = rel.Var("_int")
            return rel.RelationalAbstraction(
                tuple([x]),
                rel.atom("::std::common::int", tuple([128, v, x]))
            )
        elif (base == types.UInt128 or base == types.RowId) and isinstance(v, int):
            x = rel.Var("_uint")
            return rel.RelationalAbstraction(
                tuple([x]),
                rel.atom("::std::common::uint", tuple([128, v, x]))
            )
        else:
            return cast(Union[rel.Primitive, rel.RelationalAbstraction, rel.MetaValue, rel.Var], v)

    def handle_value_list(self, types, values) -> list[Union[rel.Primitive, rel.MetaValue, rel.RelationalAbstraction, rel.Var]]:
        result = []
        for t, v in zip(types, values):
            # splat values that are "varargs"
            if isinstance(v, tuple) and isinstance(t, ir.ListType):
                for item in v:
                    result.append(self.handle_value(t, item))
            else:
                result.append(self.handle_value(t, v))
        return result

    #
    # DATA MODEL
    #
    def handle_scalartype(self, n: ir.ScalarType):
        return n.name
    # TODO - what to generate for other kinds of types?




    #
    # TASKS
    #
    def handle_logical(self, n: ir.Logical) -> rel.Expr:
        # Generate nested expressions for a nested logical
        effects, other, aggregates, ranks = self._split_tasks(n.body)
        if effects:
            raise Exception("Cannot process nested logical with effects.")
        elif aggregates and ranks:
            raise Exception("Cannot process nested logical with both aggregates and ranks.")
        return self.generate_logical_body(other, aggregates, ranks)

    def handle_union(self, n: ir.Union) -> rel.Expr:
        # Generate nested expressions for a nested logical
        body:list[rel.Expr] = []
        for t in n.tasks:
            body.append(self.handle(t))
        return rel.Or(OrderedSet.from_iterable(body))

    def handle_rank(self, n: ir.Rank) -> rel.Expr:
        return rel.atom("rank", tuple([self.handle_var(n.result)]))

    #
    # LOGICAL QUANTIFIERS
    #
    def handle_not(self, n: ir.Not):
        return rel.Not(self.handle(n.task))

    def handle_exists(self, n: ir.Exists):
        vars = self._remove_wildcards(n.vars)
        if vars:
            return rel.Exists(
                self.handle_vars(vars),
                self.handle(n.task)
            )
        else:
            # all vars are wildcards, no need for exists
            return self.handle(n.task)

    #
    # ITERATION
    #

    #
    # RELATIONAL OPERATIONS
    #

    def handle_vars(self, vars: Tuple[ir.Var, ...]):
        return tuple([self.handle_var(v) for v in vars])

    def handle_var(self, n: ir.Var):
        name = n.name if (n.name == "_" or n.name.startswith(':')) else f"_{self._var_name(n)}"
        v = rel.Var(name)
        return self.var_map.get(v, v)

    def handle_literal(self, n: ir.Literal):
        return self.handle_value(n.type, n.value)

    def handle_data(self, n: ir.Data):
        return rel.Atom(rel.Union(tuple([self.handle_value(None, d) for d in n])), tuple([self.handle(d) for d in n.vars]))

    def generate_annotations(self, annos: Iterable[ir.Annotation]):
        """ Helper to cast the handling of ir.Annotations into a tuple of rel.Annotations. """
        filtered_annos = list(filter(lambda anno: anno.relation.name in rel_bt.builtin_annotation_names, annos))
        rel_annos = cast(Tuple[rel.Annotation, ...], self.handle_list(filtered_annos))
        return rel_annos

    def handle_annotation(self, n: ir.Annotation):
        # we know that annotations won't have vars, so we can ignore that type warning
        return rel.Annotation(
            n.relation.name,
            tuple(self.handle_value_list(self._relation_types(n.relation), n.args)) # type:ignore
        )

    def handle_update(self, n: ir.Update):
        return rel.atom(
            self._relation_name(n.relation),
            tuple(self.handle_value_list(self._relation_types(n.relation), n.args))
        )


    def handle_lookup(self, n: ir.Lookup):
        # special cases
        if n.relation == bt.cast:
            return self.handle_cast(n)
        if n.relation == bt.parse_decimal:
            return self.handle_parse_decimal(n)

        # only translate names to Rel operators if the relation is a built-in
        name = self._relation_name(n.relation)
        if bt.is_builtin(n.relation):
            name = u.rel_operator(name)
        types = self._relation_types(n.relation)
        return rel.atom(name, tuple(self.handle_value_list(types, n.args)))

    def handle_construct(self, n: ir.Construct):
        args = self.handle_value_list([None] * len(n.values), n.values) + [self.handle(n.id_var)]
        return rel.atom("rel_primitive_hash_tuple_uint128", tuple(args))


    #--------------------------------------------------
    # Built-in special cases
    #--------------------------------------------------
    def handle_cast(self, n: ir.Lookup):
        assert len(n.args) == 3
        (target_type, source, target) = n.args
        assert isinstance(target_type, ir.Type), f"Expected Type, got {type(target_type)}"
        from_type = to_type(source)
        rel_abstraction = self._convert_abs(from_type, target_type)
        types = (from_type, target_type)
        return rel.Atom(rel_abstraction, tuple(self.handle_value_list(types, (source, target))))

    def handle_parse_decimal(self, n: ir.Lookup):
        # parse_decimal in the metamodel is a binary relation, (value:String, var:DecimalType)
        # but we need to expand the lookup to (bits(var.precision), var.scale, value, var),
        # so we special case it here
        assert len(n.args) == 2
        value, var  = n.args
        assert isinstance(value, ir.Node)
        assert isinstance(var, ir.Var)
        typ = var.type
        assert isinstance(typ, ir.DecimalType)
        args = [
            rel.MetaValue(types.digits_to_bits(typ.precision)),
            rel.MetaValue(typ.scale),
            self.handle(value),
            self.handle(var)
        ]
        return rel.Atom(rel.Identifier("rel_primitive_parse_decimal"), tuple(args))

    #--------------------------------------------------
    # Helpers
    #--------------------------------------------------

    def _relation_types(self, relation: ir.Relation):
        return [f.type for f in relation.fields]

    def _relation_name(self, relation: ir.Relation):
        if helpers.is_external(relation) or helpers.builtins.is_builtin(relation):
            return relation.name
        return self.relation_name_cache.get_name(relation.id, relation.name, helpers.relation_name_prefix(relation))

    def _var_name(self, var: ir.Var):
        if var.name == "_":
            return "_"
        return self.rule_name_cache.get_name(var.id, u.sanitize(var.name.lower()))

    def _remove_wildcards(self, vars: tuple[ir.Var, ...]):
        return tuple(filter(lambda v: v.name != "_", vars))

    def _register_external_relations(self, model: ir.Model):
        # force all external relations to get a name in the cache, so that internal relations
        # cannot use those names in _relation_name
        for r in model.relations:
            if helpers.is_external(r):
                self.relation_name_cache.get_name(r.id, r.name)

    def _split_tasks(self, tasks: PySequence[ir.Task]) -> tuple[list[Union[ir.Update, ir.Output]], list[ir.Task], list[ir.Aggregate], list[ir.Rank]]:
        effects = []
        aggregates = []
        other_body = []
        ranks = []
        for task in tasks:
            if isinstance(task, (ir.Update, ir.Output)):
                effects.append(task)
            elif isinstance(task, ir.Aggregate):
                aggregates.append(task)
            elif isinstance(task, ir.Rank):
                ranks.append(task)
            else:
                other_body.append(task)
        return effects, other_body, aggregates, ranks


    def _effect_name(self, n: ir.Task):
        """ Return the name to be used for the effect (e.g. the relation name, output, etc). """
        if isinstance(n, ir.Output) and bt.export_annotation in n.annotations:
            return "Export_Relation"
        elif isinstance(n, ir.Output):
            return "output"
        elif isinstance(n, ir.Update):
            return self._relation_name(n.relation)
        else:
            raise Exception(f"Cannot retrieve effect name from node {type(n)}")

    def _effect_args(self, n: ir.Task) -> Tuple[Tuple[Any], list[ir.Task], list[rel.Expr]]:
        """
            Return the arguments for the head of an effect rule and a list of lookups to add
            to the body of the rule.

            The lookups may be necessary because Rel does not allow "missing" in the head,
            so we create a new variable, set the variable to missing in the body (the
            lookup) and use the variable in the head.

            E.g. output(None) becomes output(x): { x = missing }
        """
        orig_args = []
        handled_args = []
        if isinstance(n, ir.Output):
            args = helpers.output_values(n.aliases)
            orig_args.extend(args)
            handled_args.extend(self.handle_value_list([None] * len(args), args))
        elif isinstance(n, ir.Update):
            orig_args.extend(n.args)
            handled_args.extend(self.handle_value_list(self._relation_types(n.relation), n.args))
        else:
            raise Exception(f"Cannot retrieve effect params from node {type(n)}")

        assert len(orig_args) == len(handled_args)
        args, lookups, rel_equiv = [], [], []
        for idx, handled in enumerate(handled_args):
            if handled is None:
                var = ir.Var(types.Any, "head")
                args.append(self.handle(var))
                lookups.append(f.lookup(bt.eq, [var, orig_args[idx]]))
            elif isinstance(handled, rel.RelationalAbstraction):
                var = ir.Var(types.Any, "head")
                rel_var = self.handle(var)
                args.append(rel_var)
                rel_equiv.append(rel.create_eq(rel_var, handled))
            elif isinstance(handled, bool):
                # boolean constants need to be bound to a var in the body which is used in the head
                var = ir.Var(types.Any, "head")
                args.append(self.handle(var))
                lookups.append(f.lookup(bt.eq, [var, ir.Literal(types.Bool, handled)]))
            elif not isinstance(handled, rel.Var):
                # other constants
                args.append(handled)
            else:
                orig = orig_args[idx]
                assert isinstance(orig, ir.Var)

                # Count how many times this argument has been seen before
                cnt = handled_args[:idx].count(handled)
                if cnt == 0:
                    arg_type = self._ir_var_to_rel_type(orig)
                    args.append(rel.Var(handled.name, False, arg_type) if arg_type else handled)
                    continue

                # Deduplicate variable
                new_var = ir.Var(orig.type, handled.name + "_dup" + str(cnt))
                rel_var = self.handle_var(new_var)
                arg_type = self._ir_var_to_rel_type(orig)
                args.append(rel.Var(rel_var.name, False, arg_type) if arg_type else rel_var)
                rel_equiv.append(rel.create_eq(rel_var, handled))
        return tuple(args), lookups, rel_equiv

    def _ir_var_to_rel_type(self, v: ir.Var) -> str:
        if isinstance(v.type, ir.DecimalType):
            return u.rel_typename(v.type)
        primitive_type = to_base_primitive(v.type)
        if primitive_type:
            return u.rel_typename(primitive_type)
        elif v.type != types.Any and v.type != types.Enum:
            return "pyrel_ID"
        return ""

    def _rel_reads(self, root, reads:OrderedSet[str]|None = None) -> OrderedSet[str]:
        if reads is None:
            reads = OrderedSet()

        if isinstance(root, (tuple, list)):
            for r in root:
                self._rel_reads(r, reads)

        if isinstance(root, rel.Var):
            if root.type is not None and not root.type.startswith("::std::common::"):
                reads.add(root.type)

        elif isinstance(root, rel.Declare):
            assert isinstance(root.premise, rel.Atom)
            self._rel_reads(root.premise.args, reads)
            self._rel_reads(root.requires, reads)

        elif isinstance(root, rel.Def):
            self._rel_reads(root.params, reads)
            self._rel_reads(root.body, reads)

        elif isinstance(root, rel.Atom):
            if isinstance(root.expr, rel.Identifier):
                reads.add(root.expr.name)
            else:
                self._rel_reads(root.expr, reads)
            self._rel_reads(root.args, reads)

        elif isinstance(root, rel.RelationalAbstraction):
            self._rel_reads(root.head, reads)
            self._rel_reads(root.body, reads)

        elif isinstance(root, rel.And):
            for arg in root.body:
                self._rel_reads(arg, reads)

        elif isinstance(root, rel.Or):
            for arg in root.body:
                self._rel_reads(arg, reads)

        elif isinstance(root, rel.Exists):
            self._rel_reads(root.body, reads)

        elif isinstance(root, rel.ForAll):
            self._rel_reads(root.body, reads)

        elif isinstance(root, rel.Not):
            self._rel_reads(root.body, reads)

        elif isinstance(root, rel.BinaryExpr):
            self._rel_reads(root.lhs, reads)
            self._rel_reads(root.rhs, reads)
            reads.add(root.op)

        elif isinstance(root, rel.Product):
            for arg in root.body:
                self._rel_reads(arg, reads)

        elif isinstance(root, rel.Union):
            for arg in root.body:
                self._rel_reads(arg, reads)

        return reads



import dataclasses
import datetime
from abc import abstractmethod, ABC
from enum import Enum
from typing import Optional, TypeVar, Generic, Tuple, Callable, Sequence
from typing import Union

import relationalai.early_access.builder as qb
from relationalai.early_access.builder import std
from relationalai.early_access.builder.builder import Not, Ref, is_decimal
from relationalai.early_access.dsl.orm.relationships import Relationship, Role
from relationalai.early_access.dsl.orm.types import Concept
from relationalai.early_access.dsl.snow.common import ColumnRef, ForeignKey

_PrimitiveType = Union[int, float, str, bool, datetime.date, datetime.datetime, Enum]
FilterBy = Union[_PrimitiveType, Not, qb.Expression, tuple[qb.Expression]]

_BuiltInQbTransformer = Callable  # A callable that is a built-in QB transformer, such as string_trim, lowercase, etc.
_Transformer = Union[_BuiltInQbTransformer, qb.Relationship]
TransformWith = Union[_Transformer, Sequence[_Transformer]]

_supported_builtins_to_types = {
    std.strings.lower: (qb.String, qb.String),
    std.strings.upper: (qb.String, qb.String),
}

def _get_transform_types(obj: _Transformer):
    """Utility: return (input_type, output_type) of a transformer/relationship."""
    if isinstance(obj, qb.Relationship) and obj._arity() == 2:
        fields = getattr(obj, "_field_refs", None)
        if fields:
            return fields[0]._thing, fields[-1]._thing
    elif isinstance(obj, Callable) and obj in _supported_builtins_to_types:
        input_type, output_type = _supported_builtins_to_types[obj]
        return input_type, output_type
    raise ValueError(f"Cannot infer input/output types from {obj}. Make sure to use a valid built-in transformer or"
                     f" a binary relationship")

def _validate_transform_with(transform_with: Optional[TransformWith], input_type: Optional[qb.Concept]=None):
    if not transform_with:
        return

    elements = (transform_with,) if not isinstance(transform_with, tuple) else transform_with

    # Check the chain of types
    prev_output_type = input_type
    for idx, elem in enumerate(elements):
        if not isinstance(elem, (qb.Relationship , Callable)):
            raise ValueError(f"Expected a built-in transformer or a binary relationship, got {type(elem)}")
        input_type, output_type = _get_transform_types(elem)
        if input_type is not None and prev_output_type is not None and input_type is not prev_output_type:
            # assume decimals can be converted across each other
            if not (is_decimal(input_type) and is_decimal(prev_output_type)):
                raise ValueError(f"Type mismatch during transformation: output of #{idx-1} ({prev_output_type}) !="
                                 f" input of #{idx} ({input_type})")
        prev_output_type = output_type

# defining symbol in codegen to not expose this to the user
Symbol = Concept.builtins["Symbol"]


class BindableTable(ABC):
    """
    A class representing a bindable table.
    """

    @abstractmethod
    def key_type(self) -> qb.Concept:
        pass

    @abstractmethod
    def physical_name(self) -> str:
        pass


class BindableAttribute(ABC):

    @property
    @abstractmethod
    def table(self) -> BindableTable:
        pass

    @abstractmethod
    def physical_name(self) -> str:
        pass

    @abstractmethod
    def type(self) -> qb.Concept:
        pass

    @property
    @abstractmethod
    def references_column(self) -> Optional[ColumnRef]:
        pass

    @abstractmethod
    def column_ref(self) -> ColumnRef:
        pass

    @abstractmethod
    def decimal_scale(self) -> Optional[int]:
        pass

    @abstractmethod
    def decimal_size(self) -> Optional[int]:
        pass

    @abstractmethod
    def row_ref(self) -> Ref:
        pass


@dataclasses.dataclass(frozen=True)
class Binding:
    """
    Base class for all bindings in the model. It captures the column and the optional filter that should be applied
    to the binding.

    Examples of filters:
        - `FilterBy` can be a primitive value, such as an integer or a string, which will be used to filter the
          values in the column.
        - An expression can be used to filter the values in the column based on a more complex condition.
        - A tuple of expressions can be used to filter the values in the column based on multiple expressions, joined
          together.

        Note: all the expressions must be atomic, i.e., they cannot contain arbitrarily nested formulas.

        # binds the column to the role in the relationship, filtering out any rows that don't match the specified value
        {source}.COLUMN.binds(MyRelationship, filter_by="some_value")

        # binds the column to the role in the relationship, filtering out any rows that don't match the specified
        # expression
        {source}.COLUMN.binds(MyRelationship, filter_by=(
            {source}.ANOTHER_COLUMN >= 18
        ))

        # multiple atoms can be used too, resulting in a conjunction
        {source}.COLUMN.binds(MyRelationship[MyRole], filter_by=(
            {source}.ANOTHER_COLUMN >= 18,
            {source}.YET_ANOTHER_COLUMN == "some_value"
        ))

    Examples of transformations:
        - `TransformWith` can be a QB built-in like `strip` or `lower`.
        - A single *binary* relationship can be used to transform the values in the column, where the type of the value
          must follow the data type defined in the model for the role.
        - A tuple of relationships can be used to transform the values in the column, chaining the transformations.
          The types should chain and match, with the last one being the data type defined in the model for the role.

        # single string lowercase transformation
        {source}.COLUMN.binds(MyRelationship, transform_with=std.strings.lower)

        # chain of transformations with a strip followed by an uppercase
        {source}.COLUMN.binds(MyRelationship, transform_with=(
            std.strings.strip, std.strings.upper
        ))

        # manually defined transformation relationship
        my_transformer = Relationship("MyTransformer {String} {String}")
        s1, s2 = String.ref(), String.ref()
        define(my_transformer(s1, s2)).where(
            # ... implement the logic ...
        )

        {source}.COLUMN.binds(MyRelationship, transform_with=my_transformer)
    """
    column: 'BindableColumn'
    filter_by: Optional[FilterBy]
    transform_with: Optional[TransformWith]


@dataclasses.dataclass(frozen=True)
class RoleBinding(Binding):
    """
    RoleBinding represents a binding between a column and a role in a relationship.

    See `BindableColumn.binds` for more details on how to use this binding.
    """
    role: Role

    def __str__(self):
        return f'RoleBinding[{self.column.table.physical_name()}:{self.column.physical_name()} -> {self.role.player().name()}]'


@dataclasses.dataclass(frozen=True)
class AbstractConceptBinding(Binding):
    """
    Represents a parent class for all concept bindings. Doesn't have any specific semantics but serves as a base
    class for this class hierarchy.
    """
    entity_type: qb.Concept


@dataclasses.dataclass(frozen=True)
class IdentifierConceptBinding(AbstractConceptBinding):
    """
    Represents a binding between an identifier column and a specific entity type.

    This binding could either represent a constructor binding (instances of the entity type are constructed from the
    values), referent binding (instances of the entity type are being looked up by the values), or a subtype binding
    (instances are being looked up using the parent type's ref scheme, or it acts as a constructor for the subtype).

    See `BindableColumn.identifies` and `BindableColumn.references` for more details on how to use this binding.
    """

    def __str__(self):
        return f'IdentifierConceptBinding[{self.column.table.physical_name()}:{self.column.physical_name()} -> {self.entity_type.name()}]'


@dataclasses.dataclass(frozen=True)
class ReferentConceptBinding(AbstractConceptBinding):
    """
    Represents a binding between an identifier column and a specific entity type, where the values in the column
    are used to look up existing instances of the entity type.

    See `BindableColumn.identifies` and `BindableColumn.references` for more details on how to use this binding.
    """

    def __str__(self):
        return f'ReferentConceptBinding[{self.column.table.physical_name()}:{self.column.physical_name()} -> {self.entity_type.name()}]'


@dataclasses.dataclass(frozen=True)
class SubtypeConceptBinding(AbstractConceptBinding):
    """
    Represents a binding between an identifier column and a subtype of an entity type, where the values in the column
    are used to filter existing instances of the parent entity type. That could be done either directly (any value from
    that column is a subtype) or with the extra filter passed in `filter_by`, by filtering out the rows that do not
    match the specified value.

    See `BindableColumn.binds_subtype` for more details on how to use this binding.
    """

    def __str__(self):
        return f'SubtypeConceptBinding[{self.column.table.physical_name()}:{self.column.physical_name()} -> {self.entity_type.name()}]'


class BindableColumn(BindableAttribute, qb.Relationship, ABC):
    _source: BindableTable
    _references_column: Optional[ColumnRef]

    def __init__(self, col_name: str, col_type: qb.Concept, source: BindableTable, model):
        madlib, short_name = self._construct_name(col_name, col_type, source)
        qb.Relationship.__init__(self, madlib, model=model.qb_model(), short_name=short_name)
        self._col_name = col_name
        self._type = col_type
        self._source = source
        self._orm_model = model
        self._references_column = None

    @abstractmethod
    def _construct_name(self, col_name: str, col_type: qb.Concept, source: BindableTable) -> Tuple[str, str]:
        pass

    def identifies(self, entity_type: Concept, filter_by: Optional[FilterBy] = None,
                   transform_with: Optional[TransformWith] = None):
        """
        Binds a column that identifies an entity type. With that, the values in the column are used to construct
        instances of the entity type.

        Examples:
            {source}.ID.identifies(Person) # binds the column values to the Person entity type, constructing instances
        """
        if isinstance(entity_type, Enum):
            raise ValueError(f'Cannot bind to Enum {entity_type}, use `references` instead')
        _validate_transform_with(transform_with, input_type=self._type)
        binding = IdentifierConceptBinding(column=self, entity_type=entity_type, filter_by=filter_by,
                                           transform_with=transform_with)
        self._orm_model.binding(binding)

    def references(self, entity_type: Concept, filter_by: Optional[FilterBy] = None,
                   transform_with: Optional[TransformWith] = None):
        """
        Binds a column that references an identifier of an entity type. With that, the values in the column are
        used to look up existing instances of the entity type.

        Examples:
            {source}.ID.identifies(Person) # binds the column values to the Person entity type, constructing instances
            {another_source}.PERSON_ID.references(Person) # binds the column values to the Person entity type,
                                                          # looking up existing instances

            {source}.TRACKING_NUMER.identifies(TrackingNumber) # constructs instances of the TrackingNumber entity type
            {source}.RETURN_TRACKING_NUMBER.references(TrackingNumber) # looks up existing instances of the TrackingNumber entity type
        """
        _validate_transform_with(transform_with, input_type=self._type)
        binding = ReferentConceptBinding(column=self, entity_type=entity_type, filter_by=filter_by,
                                         transform_with=transform_with)
        self._orm_model.binding(binding)

    def binds_subtype(self, sub_type: Concept, filter_by: Optional[FilterBy] = None,
                      transform_with: Optional[TransformWith] = None):
        """
        Binds a column that contains identifier of the subtype. The values in the column are used to filter
        existing instances of the parent entity type by filtering out the rows that do not match the
        specified value.

        Note: the subtype must inherit the reference scheme of the parent entity type and not define its own. Otherwise,
        use `identifies` or `references` to bind to the subtype.

        Examples:
            # constructs instances of the CorporateAction entity type
            {source}.CORPORATE_ACTION_ID.identifies(CorporateAction)

            # partitions CorporateAction into Split subtype
            {source}.CORPORATE_ACTION_ID.binds_subtype(Split, filter_by=(
                {source}.CORPORATE_ACTION_TYPE == "SPLIT"
            ))

            # partitions CorporateAction into Merger subtype
            {source}.CORPORATE_ACTION_ID.binds_subtype(Merger, filter_by=(
                {source}.CORPORATE_ACTION_TYPE == "MERGER"
            ))
        """
        _validate_transform_with(transform_with, input_type=self._type)
        binding = SubtypeConceptBinding(column=self, entity_type=sub_type, filter_by=filter_by,
                                        transform_with=transform_with)
        self._orm_model.binding(binding)

    def binds(self, elm, filter_by: Optional[FilterBy] = None, transform_with: Optional[TransformWith] = None):
        """
        Binds the column to a role in a relationship or a role itself. It's possible to bind a column to a binary
        relationship, which will bind to the last role in the relationship.

        Examples:
            {source}.{column}.binds(MyRelationship) # binds the last role in MyRelationship

            {source}.{column}.binds(MyRelationship[MyConcept]) # binds the role played by MyConcept (if unique)

            # binding by index (may be useful for relationships with multiple roles played by the same concept)
            # BecameFriendsAt = Relationship("{Person} at {DateTime} befriended {Person}")
            friends_source.PERSON_ID.binds(BecameFriendsAt[0]) # binds the first Person role
            friends_source.BEFRIENDED_AT.binds(BecameFriendsAt[DateTime]) # binds the DateTime role
            friends_source.FRIEND_ID.binds(BecameFriendsAt[2]) # binds the second Person role (third in the relationship)
        """
        if isinstance(elm, Relationship):
            # this binds to the last role in binary relations
            if elm._arity() > 2:
                raise ValueError(f'Expected binary or unary relationship, got arity {elm._arity()}')
            roles = elm._roles()
            role = roles[-1]
        elif isinstance(elm, Role):
            role = elm
        else:
            raise Exception(
                f'Expected ORM Relationship or Role, got {type(elm)} - QB Relationships cannot be used in bindings')
        _validate_transform_with(transform_with, input_type=self._type)
        binding = RoleBinding(role=role, column=self, filter_by=filter_by, transform_with=transform_with)
        self._orm_model.binding(binding)

    @property
    def table(self):
        return self._source

    @property
    def references_column(self) -> Optional[ColumnRef]:
        return self._references_column

    @references_column.setter
    def references_column(self, ref: ColumnRef):
        self._references_column = ref

    def physical_name(self) -> str:
        return self._col_name

    def type(self) -> qb.Concept:
        return self._type

    def column_ref(self) -> ColumnRef:
        return ColumnRef(self._source.physical_name(), self.physical_name())


T = TypeVar("T", bound=BindableColumn)
class AbstractBindableTable(BindableTable, Generic[T], ABC):
    _foreign_keys: set[ForeignKey]

    def __init__(self, name: str, model, foreign_keys: set[ForeignKey]):
        super().__init__()
        self._table = name
        self._model = model
        self._cols: dict[str, T] = dict()
        self._foreign_keys = foreign_keys

    def __getattr__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]
        if key in self._cols:
            return self._cols[key]
        raise AttributeError(f'Table "{self.physical_name()}" has no column named "{key}"')

    def foreign_key(self, *refs: Tuple[BindableColumn, BindableColumn]):
        source_columns = []
        target_columns = []

        for source, target in refs:
            source_columns.append(ColumnRef(source.table.physical_name(), source.physical_name()))
            target_columns.append(ColumnRef(target.table.physical_name(), target.physical_name()))

        source_col_names = "_".join(col.column for col in source_columns)
        target_col_names = "_".join(col.column for col in target_columns)

        fk_name = f"fk_{source_col_names}__to__{target_col_names}"

        fk = ForeignKey(fk_name, source_columns, target_columns)
        self._foreign_keys.add(fk)
        self._process_foreign_key(fk)

    def key_type(self) -> qb.Concept:
        return qb.Integer

    def columns(self):
        return self._cols

    def _process_foreign_keys(self):
        for fk in self._foreign_keys:
            self._process_foreign_key(fk)

    def _process_foreign_key(self, fk):
        # TODO : this doesn't work for composite FKs
        for col in fk.source_columns:
            target_col = fk.target_columns[0]
            self._cols[col.column].references_column = target_col



from __future__ import annotations
from collections.abc import Mapping
from copy import deepcopy
import os
import time
from typing import Dict, Any, Literal
import configparser
import rich
import toml
import tomlkit
import tomlkit.exceptions

from relationalai.clients.hash_util import hash_string
from relationalai.tools.constants import FIELD_PLACEHOLDER, SNOWFLAKE_AUTHS, SNOWFLAKE_DIRECT_ACCESS_AUTHS, SNOWFLAKE_PROFILE_DEFAULTS
from relationalai.util.constants import DEFAULT_PROFILE_NAME, TOP_LEVEL_PROFILE_NAME

PUBLIC_CONFIG_KEYS = [
    "active_profile",
    "platform",
    "role",
    "rai_app_name",
    "authenticator",
    "use_graph_index",
    "reuse_model",
    "wait_for_stream_sync",
    "use_package_manager",
    "auto_suspend_mins",
    "ensure_change_tracking",
    "download_url_type",
    "use_direct_access",
]

#--------------------------------------------------
# config defaults
#--------------------------------------------------

ENV_VAR_NAME = "RAI_CONFIG_FILE"
ENDPOINT_ENV_VAR_NAME = "RAI_ENDPOINT_FILE"
DEFAULT_CONFIG_FILE = "raiconfig.toml"
CONFIG_FILE = os.getenv(ENV_VAR_NAME, DEFAULT_CONFIG_FILE)
CONFIG_FILE_SET_BY_USER = bool(os.getenv(ENV_VAR_NAME))
DEFAULT_ENDPOINT_FILE = "/tmp/relationalai/endpoints.toml"
ENDPOINT_FILE = os.getenv(ENDPOINT_ENV_VAR_NAME, DEFAULT_ENDPOINT_FILE)
LOWER_KEY_VALUES = ["platform", "authenticator"]

INTERNAL_PREFIX = "NDSOEBE"

azure_default_props = {
    "platform": "azure",
    "host": "azure.relationalai.com",
    "port": 443,
    "region": "us-east",
    "scheme": "https",
    "client_credentials_url": "https://login.relationalai.com/oauth/token",
    "client_id": FIELD_PLACEHOLDER,
    "client_secret": FIELD_PLACEHOLDER,
}

def default_engine_size(platform: Literal["snowflake"]|Literal["azure"], internal=False):
    if platform == "snowflake" and not internal:
        return SNOWFLAKE_PROFILE_DEFAULTS["engine_size"]["value"]
    else:
        return "S"

class NoDefault:
    pass
_no_default = NoDefault()

#--------------------------------------------------
# helpers
#--------------------------------------------------

def map_toml_value(value: Any):
    if value is True:
        return "true"
    if value is False:
        return "false"
    return value

class ConfigFile:
    """
    Represents of a file that contains some config information, including both raiconfig.toml and rai.config files.

    Used to handle collection and merging of configuration properties from multiple sources.
    """
    def __init__(
        self,
        path: str | None = None,
        cfg: Dict[str, Any] = {},
        *,
        format = "toml"
    ):
        if format == "toml":
            self.path:str|None = path
            self.profiles:Dict[str,Dict[str,Any]] = cfg.get("profile", {})
            self.config:Dict[str,Any] = {}
            for key in cfg:
                if key != "profile":
                    self.config[key] = cfg[key]
        elif format == "ini":
            self.path = path
            self.profiles = {
                k: {k2: v2 for k2, v2 in v.items()}
                for k, v in cfg.items()
                if k != "DEFAULT"
            }
            self.config = {}
        else:
            raise ValueError(f"Unsupported format: {format}")

    def migrate_old_keys(self):
        keymap = {
            "active-profile": "active_profile",
            "snowflake-connection": "snowflake_connection",
            "snowsql_user": "user",
            "snowsql_pwd": "password",
        }
        new_config = {}
        for key in self.config:
            new_config[keymap.get(key, key)] = self.config[key]
        self.config = new_config
        new_profiles = {}
        for profile, props in self.profiles.copy().items():
            new_profiles[profile] = {}
            for key in props.copy():
                # self.profiles[profile][keymap.get(key, key)] = props[key]
                new_profiles[profile][keymap.get(key, key)] = props[key]
        self.profiles = new_profiles

    def get_combined_profiles(self):
        combined_profiles = {}
        combined_profiles[TOP_LEVEL_PROFILE_NAME] = self.config or {}
        for profile in self.profiles:
            combined_profiles[profile] = self.profiles[profile].copy()
        return combined_profiles

    def map(self, f):
        """
        Apply a function `f` to all props dictionaries in the configuration files, including the top-level dictionary and all profiles.
        """
        combined_profiles = self.get_combined_profiles()
        for profile in combined_profiles:
            combined_profiles[profile] = f(combined_profiles[profile])
        config = combined_profiles.pop(TOP_LEVEL_PROFILE_NAME)
        return self.__class__(
            self.path, { **config, "profile": combined_profiles }
        )

    def filled_from_snowflake_connection(self):
        def fill(props):
            if "snowflake_connection" in props:
                connection_name = props.pop("snowflake_connection")
                snowflake_config = get_from_snowflake_connections_toml()
                if snowflake_config and connection_name in snowflake_config:
                    return {
                        **snowflake_config[connection_name],
                        **props,
                    }
            return props
        return self.map(fill)

    def merge(self, other):
        """
        Merge the profiles from `other` into this ConfigFile object.
        For common keys, the values from `self` take precedence.
        """
        combined_profiles = self.get_combined_profiles()
        other_combined_profiles = other.get_combined_profiles()

        all_profile_names = set(combined_profiles) | set(other_combined_profiles)

        for profile in all_profile_names:
            self_profile = combined_profiles.get(profile, {})
            other_profile = other_combined_profiles.get(profile, {})
            combined_profiles[profile] = {**other_profile, **self_profile}

        config = combined_profiles.pop(TOP_LEVEL_PROFILE_NAME, None)

        if config is not None:
            self.config = config
        self.profiles = combined_profiles


def first(x):
    return next(iter(x), None)

def _search_upwards_for_file(file:str):
    """
    Search for `file` in the current directory and all parent directories.

    Returns the absolute path to the file if found, otherwise None.
    """
    dir = os.path.abspath(os.getcwd())
    while True:
        file_path = os.path.join(dir, file)
        if os.path.isfile(file_path):
            yield file_path
        parent_dir = os.path.dirname(dir)
        if parent_dir == dir:
            break # reached the root
        dir = parent_dir

def _search_userdir_for_file(file:str):
    """
    Search for `file` in the user's home directory.

    Returns the absolute path to the file if found, otherwise None.
    """
    file_path = os.path.expanduser(f"~/{file}")
    if os.path.isfile(file_path):
        yield file_path

def _find_config_file():
    if CONFIG_FILE_SET_BY_USER and os.path.sep in CONFIG_FILE:
        path = CONFIG_FILE if os.path.isabs(CONFIG_FILE) else os.path.join(os.getcwd(), CONFIG_FILE)
        if not os.path.exists(path):
            raise FileNotFoundError(f"Config file set by environment variable but not found: {path}")
        yield path
    else:
        yield from _search_upwards_for_file(CONFIG_FILE)
        yield from _search_userdir_for_file(f".rai/{CONFIG_FILE}")
        if CONFIG_FILE != DEFAULT_CONFIG_FILE:
            yield from _search_upwards_for_file(DEFAULT_CONFIG_FILE)
            yield from _search_userdir_for_file(f".rai/{DEFAULT_CONFIG_FILE}")

def _parse_and_map_config(file:str):
    """
    Parse the config file at `file` and return a dictionary of config values.

    Handles both TOML and INI files.
    """
    if not os.path.exists(file):
        return ConfigFile(cfg={})
    if file.endswith(".toml"):
        try:
            cf = ConfigFile(file, toml.load(file))
            cf.migrate_old_keys()
            return cf
        except toml.TomlDecodeError as e:
            raise Exception(f"Error parsing {file}: {e}")
    else:
        try:
            config = configparser.ConfigParser()
            config.read(file)
            cf = ConfigFile(file, {k: v for k, v in config.items() if k != "DEFAULT"}, format="ini")
            cf.migrate_old_keys()
            return cf
        except Exception:
            # silently ignore parsing errors for INI files because
            # they are inessential to the flow
            return ConfigFile(cfg={})

def _get_full_config(profile:str|None=None) -> tuple[Dict[str,Any], str|None]:
    """
    Returns a dictionary representing the props to be used for the currently active profile. Incorporates the following rules:
    1. The `snowflake_connection` key fills in the properties from the Snowflake connections.toml file.
    2. Profiles with the same name are merged in order of appearance, with earlier profiles taking precedence.
    3. The profile indicated by the `active_profile` key is merged into other top-level properties.
    """
    files = list(_find_config_file())
    file_path = first(files)
    if not files:
        return {}, file_path
    config_files = [
        _parse_and_map_config(file).filled_from_snowflake_connection()
        for file in files
    ]
    for i in range(len(config_files) - 1, 0, -1):
        config_files[i-1].merge(config_files[i])
    if not config_files:
        return {}, file_path
    root_file = config_files[0]
    if profile:
        root_file.config["active_profile"] = profile
    if not profile and "active_profile" not in root_file.config:
        root_file.config["active_profile"] = DEFAULT_PROFILE_NAME
    if ("active_profile" in root_file.config and
        root_file.config["active_profile"] in root_file.profiles):
        active_profile = root_file.config["active_profile"]
        config = {
            **root_file.config,
            **root_file.profiles[active_profile],
        }
    else:
        config = root_file.config
    return config, file_path

def has_platform(cfg, platform):
    return any(profile.get("platform") == platform for profile in cfg.get_combined_profiles().values())

def _legacy_config_files():
    """
    Generates all legacy config files found

    Includes both ~/.rai/config and rai.config (in a parent directory
    or the user's home directory)
    """
    yield from _search_upwards_for_file("rai.config")
    yield from _search_userdir_for_file(".rai.config")
    path = os.path.expanduser("~/.rai/config")
    if os.path.exists(path):
        yield path

def legacy_config_exists():
    return any(_legacy_config_files())

def all_configs_including_legacy():
    for file in _find_config_file():
        yield _parse_and_map_config(file)
    for file in _legacy_config_files():
        yield _parse_and_map_config(file)

def get_from_snowflake_connections_toml():
    user_config_path = os.path.expanduser("~/.snowflake/connections.toml")
    if os.path.exists(user_config_path):
        try:
            snow_config = toml.load(user_config_path)
        except toml.TomlDecodeError as e:
            raise Exception(f"Error parsing {user_config_path}: {e}")
        config = {}
        for profile in snow_config:
            config[profile] = {}
            for key in snow_config[profile]:
                config[profile][key] = snow_config[profile][key]
        return config

def to_rai_config(data:Dict[str, Any]) -> Dict[str, Any]:
    from railib import config
    creds = config._read_client_credentials(data)
    _keys = ["host", "port", "region", "scheme", "audience"]
    result = {k: v for k, v in data.items() if k in _keys}
    result["credentials"] = creds
    return result


#--------------------------------------------------
# Config
#--------------------------------------------------

class ConfigStore():
    """
    Interface to the project raiconfig.toml file -- used to read and write configuration properties.
    """
    def __init__(self, file_path:str = CONFIG_FILE, toml_string:str|None = None):
        self.load(file_path, toml_string)

    def load(self, file_path:str = CONFIG_FILE, toml_string:str|None = None):
        if toml_string:
            self.file_path = "__inline__"
            self.tomldoc = tomlkit.parse(toml_string)
        else:
            self.file_path = file_path
            self.tomldoc = tomlkit.document()
            if os.path.exists(file_path):
                with open(file_path, "r") as f:
                    try:
                        content = f.read()
                        if not content.endswith("\n"):
                            content += "\n"
                        self.tomldoc = tomlkit.parse(content)
                    except toml.TomlDecodeError as e:
                        raise Exception(f"Error parsing {file_path}: {e}")

    def get(self, key, default=None):
        return self.tomldoc.get(key, default)

    def set(self, key, value):
        self.tomldoc[key] = value

    def items(self):
        return self.tomldoc.items()

    def get_profiles(self):
        return self.get("profile", {})

    def num_profiles(self):
        return len(self.get_profiles())

    def switch_active_profile(self, profile:str):
        self.tomldoc["active_profile"] = profile
        self.save()

    def with_new_profile(self, config) -> tomlkit.TOMLDocument:
        document = deepcopy(self.tomldoc)
        profiles = document.pop("profile", None)
        if not profiles:
            profiles = tomlkit.table()
        table_for_new_profile = tomlkit.table()
        for key, value in config.items_with_dots():
            toml_key = tomlkit.key(key.split(".") if "." in key else key)
            table_for_new_profile.add(toml_key, value)

        new_profiles_table = tomlkit.table(is_super_table=True)
        new_profiles_table.add(config.profile, table_for_new_profile)
        for k, v in profiles.items():
            if k == config.profile:
                continue
            new_profiles_table.add(k, v)
        document.add("profile", new_profiles_table)

        return document

    def add_profile(self, config):
        self.tomldoc = self.with_new_profile(config)

    def save(self):
        try:
            os.makedirs(os.path.dirname(self.file_path), exist_ok=True)
        except OSError:
            pass
        try:
            with open(self.file_path, "w") as f:
                f.write(self.tomldoc.as_string())
        except Exception:
            raise Exception(f"Error saving the config to {self.file_path}.")

    def change_active_profile(self, profile:str):
        document = tomlkit.document()
        document.add("active_profile", profile) # type: ignore
        document.add(tomlkit.nl())
        for key, value in self.tomldoc.items():
            if key != "active_profile":
                document.add(key, value) # type: ignore
        self.tomldoc = document

    def __str__(self):
        return self.tomldoc.as_string()

    def get_config(self):
        props = {}
        for key, value in self.items():
            if key != "profile":
                props[key] = value
        profile = None
        if "active_profile" in props:
            active_profile = props.pop("active_profile")
            profiles = self.get("profile", {})
            if active_profile in profiles:
                props.update(profiles[active_profile])
                profile = active_profile
        config = Config(props, fetch=False)
        if profile:
            config.profile = profile
        return config

class Config():
    """
    Represents an active collection of configuration properties.
    """
    def __init__(self, profile:str|Dict[str,Any]|None=None, fetch=True):
        supplied_props = None
        if isinstance(profile, dict):
            supplied_props = profile
            profile = TOP_LEVEL_PROFILE_NAME
        if fetch:
            self.fetch(profile)
        else:
            self.profile = TOP_LEVEL_PROFILE_NAME
            self.props = {}
            self.file_path = None
        if supplied_props is not None:
            for k, v in supplied_props.items():
                self.set(k, v)
            if not self.file_path:
                self.file_path = "__inline__"
        self._handle_snowflake_fallback_configurations()

    def fetch(self, profile:str|None=None):
        from relationalai.environments import runtime_env, TerminalEnvironment
        profile = profile or os.environ.get("RAI_PROFILE", None)
        cfg, path = _get_full_config(profile if profile != TOP_LEVEL_PROFILE_NAME else None)
        cfg_profile = cfg.get("active_profile", TOP_LEVEL_PROFILE_NAME)

        self.profile = profile or cfg_profile

        # Handle the case where the config does not exist and we are in a terminal environment
        if profile is None and isinstance(runtime_env, TerminalEnvironment):
            from relationalai.errors import NonExistentConfigFileError, InvalidActiveProfileError
            if cfg == {}:
                raise NonExistentConfigFileError(DEFAULT_CONFIG_FILE)

            # Handle the scenario where the active profile is set to invalid profile name
            if set(cfg.keys()) == {"active_profile"}:
                raise InvalidActiveProfileError(cfg.get("active_profile", "unknown"))

        self.file_path = path
        self.props = cfg

    def clone_profile(self):
        self.props = {k: v for k, v in self.props.items()}

    def clone(self):
        return deepcopy(self)

    def get(self, name:str, default:Any=_no_default, strict:bool=True):
        parts = name.split(".")
        props = self.props

        for part in parts[:-1]:
            props = props.get(part)
            if props is None:
                props = {}
                break

        val = props.get(parts[-1], os.environ.get(name, default))
        if val is _no_default and strict:
            raise Exception(f"Missing config value for '{name}'")
        if val is _no_default:
            val = None
        if name in LOWER_KEY_VALUES and isinstance(val, str):
            return val.lower()
        return val

    def get_bool(self, name: str, default:bool|NoDefault=_no_default) -> bool:
        got = self.get(name, default, True)
        assert isinstance(got, bool), f"Config field {name} is set to an invalid value. Got {type(got)} '{got}' but expected a value of type 'bool'"
        return got

    def set(self, name:str, value:str|int):
        parts = name.split(".")
        props = self.props
        for part in parts[:-1]:
            if part not in props:
                props[part] = {}
            props = props[part]
        props[parts[-1]] = value

    def clear_props(self):
        self.props = {}

    def unset(self, name:str):
        del self.props[name]

    def update(self, other):
        self.props.update(other)

    def __contains__(self, key):
        return key in self.props

    def items(self):
        return self.props.items()

    def items_with_dots(self):
        return items_with_dots(self.props)

    def __str__(self):
        return "\n".join(f"{k}: {map_toml_value(v)}" for k, v in self.items_with_dots())

    def get_hash(self):
        user = self.get("user") if self.get("platform") == "snowflake" else self.get("client_id")
        return hash_string((user or "") + str(time.time_ns()))

    def to_rai_config(self) -> Dict[str, Any]:
        return to_rai_config(self.props)
    
    def _handle_snowflake_fallback_configurations(self):
        if not self.get("platform", "") == "snowflake":
            return
        if self.get("use_direct_access", False):
            if str(self.get("authenticator", "")).lower() not in SNOWFLAKE_DIRECT_ACCESS_AUTHS:
                # import here to avoid circular imports
                from relationalai.debugging import warn
                from relationalai.errors import DirectAccessInvalidAuthWarning
                warn(DirectAccessInvalidAuthWarning(str(self.get("authenticator", "")), SNOWFLAKE_DIRECT_ACCESS_AUTHS.keys()))
                self.set("use_direct_access", False)

    def _fill_in_with_defaults(self, defaults: Dict[str, Any], **kwargs):
        props = {k: v for k, v in kwargs.items() if k in defaults}
        self.update({
            **defaults,
            **self.props,
            **props,
        })

    def fill_in_with_azure_defaults(self, **kwargs):
        self._fill_in_with_defaults(azure_default_props, **kwargs)

    def fill_in_with_snowflake_defaults(self, **kwargs):
        authenticator = kwargs.get("authenticator", self.get("authenticator")) or "snowflake"
        authenticator = authenticator.lower()
        defaults = {k: v["value"] for k, v in SNOWFLAKE_AUTHS[authenticator].items()}
        self._fill_in_with_defaults(defaults, **kwargs)

    def fill_in_with_defaults(self):
        platform = self.get("platform", None)
        if platform == "azure":
            self.fill_in_with_azure_defaults()
        elif platform == "snowflake":
            self.fill_in_with_snowflake_defaults()
        else:
            self.set("platform", "snowflake")
            self.fill_in_with_snowflake_defaults()

    def is_internal_account(self):
        acct = self.get("account", "")
        acct = acct if isinstance(acct, str) else ""
        return acct.upper().startswith(INTERNAL_PREFIX)

    def get_default_engine_size(self):
        config_size = self.get("engine_size", None)
        if config_size is None:
            return default_engine_size("snowflake", internal=self.is_internal_account())
        else:
            return config_size

    def get_default_auto_suspend_mins(self):
        auto_suspend_mins = self.get("auto_suspend_mins", None)
        if auto_suspend_mins is not None:
            return int(auto_suspend_mins)
        else:
            return None

    def set_default_engine_size(self):
        self.set("engine_size", default_engine_size("snowflake", internal=self.is_internal_account()))

    def show_all_engine_sizes(self):
        return self.get("show_all_engine_sizes", None) or self.is_internal_account()

    def get_data_freshness_mins(self):
        freshness = self.get("data_freshness_mins", None)
        if freshness is None:
            return None
        return int(freshness)

#--------------------------------------------------
# helpers
#--------------------------------------------------

def items_with_dots(d: dict):
    for k, v in d.items():
        if isinstance(v, dict):
            for sub_k, sub_v in items_with_dots(v):
                yield f"{k}.{sub_k}", sub_v
        else:
            yield k, v

def save_config(toml_string: str, overwrite=False):
    """
    Convenience function to save a raiconfig.toml file with the given toml_string.

    rai.save_config(\"\"\"
    auto_suspend_mins = 30
    \"\"\")
    """
    if not overwrite and os.path.exists(CONFIG_FILE):
        raise ValueError(f"{CONFIG_FILE} already exists. Use overwrite=True to overwrite it.")
    try:
        doc = tomlkit.parse(toml_string)
    except tomlkit.exceptions.TOMLKitError as e:
        raise ValueError(f"Invalid TOML string: {e}")
    if 'profile' not in doc:
        prefix = "active_profile = \"default\"\n\n[profile.default]\n"
        toml_string = prefix + toml_string
    elif 'active_profile' not in doc:
        full_profile = doc['profile']
        assert isinstance(full_profile, Mapping), "Profile should be a table"
        profile = next(iter(full_profile.keys()))
        prefix = f"active_profile = \"{profile}\"\n\n"
        toml_string = prefix + toml_string
    with open(CONFIG_FILE, "w") as f:
        f.write(toml_string)
    rich.print(f"[green]✓ {CONFIG_FILE} saved!")



from __future__ import annotations
from typing import Union

from .. import builder as b
from .std import _String, _Integer, _make_expr

# TODO(coey) can we allow multiple args to this, and convert each to string and concatenate? similar to e.g. julia `string("foo", 1, 'b')`. then we don't need a separate concat function (and concat only works on strings, which is inconvenient).
def string(s: Union[b.Producer, str, float, int]) -> b.Expression:
    return _make_expr("string", s, b.String.ref("res"))

def concat(s0: _String, s1: _String, *args: _String) -> b.Expression:
    res = _make_expr("concat", s0, s1, b.String.ref("res0"))
    for i, s in enumerate(args):
        res = _make_expr("concat", res, s, b.String.ref(f"res{i+1}"))
    return res

def len(s: _String) -> b.Expression:
    return _make_expr("num_chars", s, b.Int128.ref("res"))

def startswith(s0: _String, s1: _String) -> b.Expression:
    return _make_expr("starts_with", s0, s1)

def endswith(s0: _String, s1: _String) -> b.Expression:
    return _make_expr("ends_with", s0, s1)

def contains(s0: _String, s1: _String) -> b.Expression:
    return _make_expr("contains", s0, s1)

def substring(s: _String, start: _Integer, stop: _Integer) -> b.Expression:
    # unlike Python, Rel's range is 1..stop inclusive
    return _make_expr("substring", s, start+1, stop, b.String.ref("res"))

def like(s: _String, pattern: _String) -> b.Expression:
    return _make_expr("like_match", s, pattern)

def lower(s: _String) -> b.Expression:
    return _make_expr("lower", s, b.String.ref("res"))

def upper(s: _String) -> b.Expression:
    return _make_expr("upper", s, b.String.ref("res"))

def strip(s: _String) -> b.Expression:
    return _make_expr("strip", s, b.String.ref("res"))



from abc import ABC, abstractmethod
from typing import List

from relationalai.early_access.dsl.core.utils import generate_stable_uuid
from relationalai.early_access.dsl.ontologies.roles import AbstractRole


class Constraint(ABC):
    @abstractmethod
    def validate(self, ontology) -> bool:
        """Validate the constraint against the ontology."""

    @abstractmethod
    def _unique_name(self) -> str:
        """Validate the constraint against the ontology."""

    @abstractmethod
    def roles(self) -> list[AbstractRole]:
        """Return the roles of the constraint."""

    def guid(self):
        return generate_stable_uuid(self._unique_name())

    def __eq__(self, other):
        if isinstance(other, Constraint):
            return self.guid() == other.guid()
        return False

    def __hash__(self):
        return hash(self.guid())


class Unique(Constraint):
    def __init__(self, *roles, is_preferred_identifier=False):
        self._roles = list(roles)
        self.is_preferred_identifier = is_preferred_identifier

    def roles(self):
        return self._roles

    def validate(self, ontology) -> bool:
        # Implement uniqueness validation logic here
        return True

    def _unique_name(self) -> str:
        return f'Unique{"".join(role.guid() for role in self._roles)}'

    def __repr__(self):
        return f'Unique(({", ".join(role.ref_name() for role in self._roles)}), preferred_identifier={self.is_preferred_identifier})'

class Mandatory(Constraint):
    def __init__(self, role: AbstractRole):
        self.role = role

    def roles(self):
        return [self.role]

    def validate(self, ontology) -> bool:
        return True

    def _unique_name(self) -> str:
        return f'Mandatory{self.role.guid()}'

    def __repr__(self):
        return f'Mandatory({self.role.ref_name()})'

class RoleValueConstraint(Constraint):
    def __init__(self, role: AbstractRole, values: List[str]):
        self._role = role
        self._values = values

    def roles(self):
        return [self._role]

    def values(self):
        return self._values

    def validate(self, ontology) -> bool:
        return True

    def _unique_name(self) -> str:
        return f'RoleValueConstraint{self._role.guid()}'

    def role(self):
        return self._role

    def __repr__(self):
        return f'RoleValueConstraint({self._role.ref_name()}, values={self._values})'



#--------------------------------------------------
# Compiler
#--------------------------------------------------

from relationalai.early_access.lqp import ir as lqp, model2lqp
from relationalai.early_access.lqp.passes import lqp_passes
from relationalai.early_access.metamodel import ir, compiler as c

from typing import Optional

class Compiler(c.Compiler):
    def __init__(self):
        super().__init__(lqp_passes())

    def do_compile(self, model: ir.Model, options:dict={}) -> tuple[Optional[tuple], lqp.Transaction]:
        fragment_id: bytes = options.get("fragment_id", bytes(404))
        return model2lqp.to_lqp(model, fragment_id)



from typing import Any, Union

import relationalai.early_access.builder as qb
from relationalai.early_access.builder.builder import python_types_to_concepts
from relationalai.early_access.dsl.core.utils import generate_stable_uuid
from relationalai.early_access.dsl.orm.relationships import Relationship


class Concept(qb.Concept):

    def __init__(self, model, name: str, extends: list[Any] = [], identify_by:dict[str, Any]={}):
        self._dsl_model = model
        # create an orm Relationship for each Concept to be able to refer to them in DSL model
        identify_args = {}
        if identify_by:
            for k, v in identify_by.items():
                if python_types_to_concepts.get(v):
                    v = python_types_to_concepts[v]
                if isinstance(v, qb.Concept):
                    identify_args[k] = Relationship(self._dsl_model, f"{{{name}}} has {{{k}:{v._name}}}", short_name=k)
                elif isinstance(v, type) and issubclass(v, self._dsl_model.Enum): #type: ignore
                    identify_args[k] = Relationship(self._dsl_model, f"{{{name}}} has {{{k}:{v._concept._name}}}", short_name=k)
                elif isinstance(v, qb.Relationship):
                    identify_args[k] = v
                else:
                    raise ValueError(f"identify_by must be either a Concept or Relationship: {k}={v}")
        super().__init__(name, extends, model.qb_model(), identify_args)
        self._dsl_model._add_concept(self)
        self._dsl_model.reasoner().new_concept(self)
        if identify_by:
            # once concept is added, add constraints required for reference schema
            self._dsl_model._ref_scheme_constraints(*identify_args.values())

    def identify_by(self, *relations:Union[qb.Relationship, qb.RelationshipReading]):
        super().identify_by(*relations)
        self._dsl_model._ref_scheme_constraints(*relations)

    def __repr__(self):
        return f"Concept({self._name})"

    def _guid(self):
        return generate_stable_uuid(f"{self._id}")



from relationalai.early_access.builder import annotations
from relationalai.early_access.dsl.bindings.common import BindableColumn, AbstractBindableTable, BindableTable, Symbol
from relationalai.early_access.dsl.snow.common import ColumnMetadata, _map_rai_type

import relationalai.early_access.builder as qb


#=
# Bindable classes and interfaces.
#=

class BindableSnowflakeColumn(BindableColumn):
    _metadata: ColumnMetadata

    def __init__(self, metadata: ColumnMetadata, table: 'SnowflakeTable', model):
        col_name = metadata.name
        col_type = _map_rai_type(metadata)
        super().__init__(col_name, col_type, table, model)
        self._metadata = metadata
        self.annotate(annotations.external)  # so we generate no declares, as there's one for the table

    def _construct_name(self, col_name: str, col_type: qb.Concept, source: BindableTable):
        """
        Handles the construction of the Relationship name components.

        For tabular sources coming from Snowflake, we need to specialize the relation that's been created for the table.
        We use separate QB relationships, as this allows proper typing, but in core Rel it's treated as a single
        relation; hence the name is the same.
        """
        short_name = source.physical_name()
        madlib = f"{short_name} {{col_name:Symbol}} {{row_id:RowId}} {{{col_type}}}"
        return madlib, short_name

    def __call__(self, *args):
        """
        Allows the column to be called as a function, which is handy for manual data weaving.

        Example:
            row = Integer.ref()
            where(
                {source}.ID(row, id),
                {source}.NAME(row, name),
                person := Person.new(id=id)
            ).define(
                person,
                Person.name(person, name)
            )
        """
        if len(args) == 2:
            args = [Symbol(self.physical_name()), *args]
        if len(args) != 3:
            raise ValueError(f'Expected 2 arguments passed to a call to BindableColumn, got {len(args)}')
        return qb.Relationship.__call__(self, *args)

    @property
    def metadata(self):
        return self._metadata

    def row_ref(self):
        assert len(self._field_refs) == 3, 'Snowflake Column relation must have exactly three field references'
        row_ref = self._field_refs[1]
        assert row_ref._thing._name == 'RowId', 'First field must be of type RowId'
        return row_ref

    def decimal_scale(self):
        return self._metadata.numeric_scale

    def decimal_size(self):
        precision = self._metadata.numeric_precision
        if precision is not None:
            if 1 <= precision <= 18:
                return 64
            elif 18 < precision <= 38:
                return 128
            raise ValueError(f'Precision {precision} is not supported (max: 38)')
        return precision

    def __repr__(self):
        return f"Snowflake:{super().__repr__()}"


class SnowflakeTable(AbstractBindableTable[BindableSnowflakeColumn]):

    def __init__(self, fqn: str, model):
        self._metadata = model.api().table_metadata(fqn)
        super().__init__(fqn, model, self._metadata.foreign_keys)
        self._initialize(model)

    def _initialize(self, model):
        self._model = model
        self._cols = {col.name: BindableSnowflakeColumn(col, self, model) for col in self._metadata.columns}
        self._process_foreign_keys()
        # initialize the table so that the graph index can be updated
        self._initialize_qb_table()
        self._generate_declare()

    def _initialize_qb_table(self):
        from relationalai.early_access.builder.snowflake import Table as QBTable
        self._qb_table = QBTable(self._table)
        QBTable._used_sources.add(self._qb_table)

    def _generate_declare(self):
        src = f"declare {self.physical_name()}"
        self._model.qb_model().define(qb.RawSource('rel', src))

    def __str__(self):
        # returns the name of the table, as well as the columns and their types
        return self.physical_name() + ':\n' + '\n'.join(
            [f' {col.metadata.name} {col.metadata.datatype}' for _, col in self._cols.items()]
        ) + '\n' + '\n'.join(
            [f' {fk.source_columns} -> {fk.target_columns}' for fk in self._foreign_keys]
        )

    def physical_name(self):
        # physical relation name is always in the form of `{database}_{schema}_{table}
        return f"{self._table.lower()}".replace('.', '_')



from typing import cast

from relationalai.early_access.metamodel import ir, compiler as c, factory as f, builtins
from relationalai.early_access.metamodel.builtins import concept_relation_annotation
from relationalai.early_access.metamodel.util import ordered_set, OrderedSet
from relationalai.early_access.metamodel.dependency import analyze_bindings


class SortOutputQuery(c.Pass):
    """
        This pass reorders lookups and logical tasks inside the model to ensure deterministic ordering of dependencies
        for queries that produce output.

        Algorithm Overview:
        -------------------
        1. For each top-level logical in the model that contains `ir.Output` inside:
           - Partition its inner tasks into:
             * Logical nodes
             * Top-level lookups (non-builtins)
             * Other tasks (including builtins)

        2. Build a mapping from each variable to the top-level lookups that produce it.

        3. For each logical node:
           a. Identify bound variables (variables it consumes).
           b. Sort **only the inner lookups** (excluding builtins) in a deterministic order
              based on their variable dependencies.
           c. Append any non-lookup or builtin tasks in their original order.
           d. Find the "best" top-level lookup that matches the logical's input variables,
              preferring lookups without `concept_relation_annotation` in their annotations.
              If found, add it to the new top-level body and remove it from the pool so
              it won't be reused.

        4. After all logical nodes are processed:
           - Add any remaining reachable top-level lookups by expanding the current variable
             domain.
           - Append any leftover unconnected lookups.
           - Append the other inner tasks (builtins, constructs, etc.).

        5. Replace the original logical body with the reordered one.

        Stability:
        ----------
        - Applies only after Flatten pass.
        - Only logicals containing `ir.Output` are processed, so non-output queries remain untouched.

        Examples:

            #1

            Flatten IR:
                Logical
                    Logical
                        Person(person::Person)
                        Logical ^[car=None]
                            Car(car::Car)
                            owns(person::Person, car::Car)
                        Car(car::Car)
                        Logical ^[state=None]
                            State(state::State)
                            registered_in(car::Car, state::State)
                        -> output[person::Person, car::Car, state::State](car::Car as 'car', state::State as 'state')

            Sorted IR:
                Logical
                    Logical
                        Person(person::Person)
                        Logical ^[car=None]
                            owns(person::Person, car::Car)
                            Car(car::Car)
                        Car(car::Car)
                        Logical ^[state=None]
                            registered_in(car::Car, state::State)
                            State(state::State)
                        -> output[person::Person, car::Car, state::State](car::Car as 'car', state::State as 'state')

            #2

            Flatten IR:
                Logical
                    Logical
                        Student(student::Student)
                        goes_at(student::Student, school::School)
                        subject(school::School, subject::Subject)
                        desc(subject::Subject, desc::String)
                        desc::String = "English"
                        Logical ^[id=None]
                            id(student::Student, id::Int128)
                        Logical ^[name=None, course=None]
                            attends(student::Student, course::Course)
                            instructor(course::Course, instructor::Instructor)
                            name(instructor::Instructor, name::String)
                        -> output[student::Student, course::Course, subject::Subject](id::Int128 as 'id', name::String as 'name', desc::String as 'desc')

            Sorted IR:
                Logical
                    Logical
                        goes_at(student::Student, school::School)
                        Logical ^[id=None]
                            id(student::Student, id::Int128)
                        Student(student::Student)
                        Logical ^[name=None, course=None]
                            attends(student::Student, course::Course)
                            instructor(course::Course, instructor::Instructor)
                            name(instructor::Instructor, name::String)
                        subject(school::School, subject::Subject)
                        desc(subject::Subject, desc::String)
                        desc::String = "English"
                        -> output[student::Student, course::Course, subject::Subject](id::Int128 as 'id', name::String as 'name', desc::String as 'desc')
    """
    def rewrite(self, model: ir.Model, options: dict = {}) -> ir.Model:
        root_logical = cast(ir.Logical, model.root)

        output_task_ids = set()
        for task in root_logical.body:
            if isinstance(task, ir.Logical) and self._is_query(task):
                output_task_ids.add(task.id)

        if not output_task_ids:
            return model

        new_model_body = []
        bindings_info = analyze_bindings(root_logical)
        for task in root_logical.body:
            if task.id not in output_task_ids:
                new_model_body.append(task)
                continue

            id_to_logical = {}
            id_to_top_level_lookup = {}
            other_inner_tasks = []

            top_level_logical = cast(ir.Logical, task)

            # Partition inner tasks
            for inner_task in top_level_logical.body:
                if isinstance(inner_task, ir.Logical):
                    id_to_logical[inner_task.id] = inner_task
                elif isinstance(inner_task, ir.Lookup):
                    if builtins.is_builtin(inner_task.relation):
                        other_inner_tasks.append(inner_task)
                    else:
                        id_to_top_level_lookup[inner_task.id] = inner_task
                else:
                    other_inner_tasks.append(inner_task)

            # Variable → top-level lookups producing it
            var_to_top_lookups = {}
            for lookup in id_to_top_level_lookup.values():
                for var in bindings_info.output_bindings.get(lookup.id, ordered_set()):
                    var_to_top_lookups.setdefault(var, []).append(lookup)

            # --- Process logicals and directly attach best matching top-level lookup ---
            new_top_level_body = []
            domain = ordered_set()

            for lid, logical in id_to_logical.items():
                original_bound_vars = bindings_info.input_bindings.get(logical.id, ordered_set())
                bound_vars = OrderedSet.from_iterable(original_bound_vars)
                domain.update(bound_vars)

                # Order inner lookups in a single pass
                ordered_inner, leftover_inner = self._order_lookups(
                    [task for task in logical.body if isinstance(task, ir.Lookup) and not builtins.is_builtin(task.relation)],
                    bindings_info.output_bindings,
                    bound_vars
                )

                ordered_inner.extend(t for t in logical.body if not isinstance(t, ir.Lookup) or
                                     (isinstance(t, ir.Lookup) and builtins.is_builtin(t.relation)))
                sorted_logical = f.logical(ordered_inner + leftover_inner, logical.hoisted, logical.engine, logical.annotations)

                # Find best candidate top-level lookup using var_to_top_lookups
                candidates = OrderedSet.from_iterable(c for var in original_bound_vars for c in reversed(var_to_top_lookups.get(var, ())))
                best = self._pick_best_candidate(candidates)
                if best:
                    new_top_level_body.append(best)
                    outputs = bindings_info.output_bindings.get(best.id, ordered_set())
                    domain.update(outputs)
                    # remove from mapping so it won't be picked again
                    for var in outputs:
                        var_to_top_lookups[var] = [lookup for lookup in var_to_top_lookups[var] if lookup != best]

                new_top_level_body.append(sorted_logical)

            # --- Add any remaining top-level lookups reachable from domain ---
            reachable = ordered_set()
            queue = OrderedSet.from_iterable(domain)
            while queue:
                var = queue.pop()
                for lookup in var_to_top_lookups.get(var, ()):
                    if lookup not in reachable:
                        reachable.add(lookup)
                        new_top_level_body.append(lookup)
                        new_vars = bindings_info.output_bindings.get(lookup.id, ordered_set())
                        domain.update(new_vars)
                        queue.update(new_vars)

            # Add any leftover unconnected lookups
            unused = [lookup for lookup in id_to_top_level_lookup.values() if lookup not in new_top_level_body]
            new_top_level_body.extend(unused)
            new_top_level_body.extend(other_inner_tasks)

            new_model_body.append(
                f.logical(new_top_level_body, top_level_logical.hoisted, top_level_logical.engine,
                          list(top_level_logical.annotations))
            )

        return model.reconstruct(model.engines, model.relations, model.types, f.logical(new_model_body),
                                 model.annotations)

    def _order_lookups(self, lookups, outputs_map, bound_vars):
        ordered = []
        queue = lookups[:]
        while queue:
            progress = False
            remaining = []
            for lookup in queue:
                outputs = outputs_map.get(lookup.id, ordered_set())
                if outputs & bound_vars:
                    ordered.append(lookup)
                    bound_vars.update(outputs)
                    progress = True
                else:
                    remaining.append(lookup)
            if not progress:
                break
            queue = remaining
        return ordered, queue

    def _pick_best_candidate(self, candidates):
        if not candidates:
            return None
        best = None
        for candidate in candidates:
            if not best:
                best = candidate
            elif concept_relation_annotation not in candidate.relation.annotations:
                best = candidate
        return best

    def _is_query(self, task: ir.Logical):
        for sub_task in task.body:
            if isinstance(sub_task, ir.Output):
                return True
        return False